from enum import Enum
from typing import List, Literal, get_args


class ActorEventType(str, Enum):
    """Possible values of actor event type."""

    #: Info about resource usage of the actor
    SYSTEM_INFO = 'systemInfo'
    #: Sent when the actor is about to migrate
    MIGRATING = 'migrating'
    #: Sent when the actor should persist its state (every minute or when migrating)
    PERSIST_STATE = 'persistState'
    #: Sent when the actor is aborting
    ABORTING = 'aborting'


class ApifyEnvVars(str, Enum):
    """Possible Apify env vars."""

    ACT_ID = 'APIFY_ACT_ID'
    ACT_RUN_ID = 'APIFY_ACT_RUN_ID'
    ACTOR_BUILD_ID = 'APIFY_ACTOR_BUILD_ID'
    ACTOR_BUILD_NUMBER = 'APIFY_ACTOR_BUILD_NUMBER'
    ACTOR_EVENTS_WS_URL = 'APIFY_ACTOR_EVENTS_WS_URL'
    ACTOR_ID = 'APIFY_ACTOR_ID'
    ACTOR_RUN_ID = 'APIFY_ACTOR_RUN_ID'
    ACTOR_TASK_ID = 'APIFY_ACTOR_TASK_ID'
    API_BASE_URL = 'APIFY_API_BASE_URL'
    API_PUBLIC_BASE_URL = 'APIFY_API_PUBLIC_BASE_URL'
    CHROME_EXECUTABLE_PATH = 'APIFY_CHROME_EXECUTABLE_PATH'
    CONTAINER_PORT = 'APIFY_CONTAINER_PORT'
    CONTAINER_URL = 'APIFY_CONTAINER_URL'
    DEDICATED_CPUS = 'APIFY_DEDICATED_CPUS'
    DEFAULT_BROWSER_PATH = 'APIFY_DEFAULT_BROWSER_PATH'
    DEFAULT_DATASET_ID = 'APIFY_DEFAULT_DATASET_ID'
    DEFAULT_KEY_VALUE_STORE_ID = 'APIFY_DEFAULT_KEY_VALUE_STORE_ID'
    DEFAULT_REQUEST_QUEUE_ID = 'APIFY_DEFAULT_REQUEST_QUEUE_ID'
    DISABLE_BROWSER_SANDBOX = 'APIFY_DISABLE_BROWSER_SANDBOX'
    DISABLE_OUTDATED_WARNING = 'APIFY_DISABLE_OUTDATED_WARNING'
    FACT = 'APIFY_FACT'
    HEADLESS = 'APIFY_HEADLESS'
    INPUT_KEY = 'APIFY_INPUT_KEY'
    INPUT_SECRETS_PRIVATE_KEY_FILE = 'APIFY_INPUT_SECRETS_PRIVATE_KEY_FILE'
    INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE = 'APIFY_INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE'
    IS_AT_HOME = 'APIFY_IS_AT_HOME'
    LOCAL_STORAGE_DIR = 'APIFY_LOCAL_STORAGE_DIR'
    LOG_FORMAT = 'APIFY_LOG_FORMAT'
    LOG_LEVEL = 'APIFY_LOG_LEVEL'
    MEMORY_MBYTES = 'APIFY_MEMORY_MBYTES'
    META_ORIGIN = 'APIFY_META_ORIGIN'
    PERSIST_STORAGE = 'APIFY_PERSIST_STORAGE'
    PROXY_HOSTNAME = 'APIFY_PROXY_HOSTNAME'
    PROXY_PASSWORD = 'APIFY_PROXY_PASSWORD'
    PROXY_PORT = 'APIFY_PROXY_PORT'
    PROXY_STATUS_URL = 'APIFY_PROXY_STATUS_URL'
    SDK_LATEST_VERSION = 'APIFY_SDK_LATEST_VERSION'
    STARTED_AT = 'APIFY_STARTED_AT'
    TIMEOUT_AT = 'APIFY_TIMEOUT_AT'
    TOKEN = 'APIFY_TOKEN'
    USER_ID = 'APIFY_USER_ID'
    WORKFLOW_KEY = 'APIFY_WORKFLOW_KEY'
    XVFB = 'APIFY_XVFB'

    # Extra ones not in @apify/consts:
    METAMORPH_AFTER_SLEEP_MILLIS = 'APIFY_METAMORPH_AFTER_SLEEP_MILLIS'
    PERSIST_STATE_INTERVAL_MILLIS = 'APIFY_PERSIST_STATE_INTERVAL_MILLIS'
    PURGE_ON_START = 'APIFY_PURGE_ON_START'


_INTEGER_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.CONTAINER_PORT,
    ApifyEnvVars.DEDICATED_CPUS,
    ApifyEnvVars.LOG_LEVEL,
    ApifyEnvVars.MEMORY_MBYTES,
    ApifyEnvVars.METAMORPH_AFTER_SLEEP_MILLIS,
    ApifyEnvVars.PERSIST_STATE_INTERVAL_MILLIS,
    ApifyEnvVars.PROXY_PORT,
]

INTEGER_ENV_VARS: List[_INTEGER_ENV_VARS_TYPE] = list(get_args(_INTEGER_ENV_VARS_TYPE))

_BOOL_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.DISABLE_BROWSER_SANDBOX,
    ApifyEnvVars.DISABLE_OUTDATED_WARNING,
    ApifyEnvVars.HEADLESS,
    ApifyEnvVars.IS_AT_HOME,
    ApifyEnvVars.PERSIST_STORAGE,
    ApifyEnvVars.PURGE_ON_START,
    ApifyEnvVars.XVFB,
]

BOOL_ENV_VARS: List[_BOOL_ENV_VARS_TYPE] = list(get_args(_BOOL_ENV_VARS_TYPE))

_DATETIME_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.STARTED_AT,
    ApifyEnvVars.TIMEOUT_AT,
]

DATETIME_ENV_VARS: List[_DATETIME_ENV_VARS_TYPE] = list(get_args(_DATETIME_ENV_VARS_TYPE))

_STRING_ENV_VARS_TYPE = Literal[
    ApifyEnvVars.ACT_ID,
    ApifyEnvVars.ACT_RUN_ID,
    ApifyEnvVars.ACTOR_BUILD_ID,
    ApifyEnvVars.ACTOR_BUILD_NUMBER,
    ApifyEnvVars.ACTOR_EVENTS_WS_URL,
    ApifyEnvVars.ACTOR_ID,
    ApifyEnvVars.ACTOR_RUN_ID,
    ApifyEnvVars.ACTOR_TASK_ID,
    ApifyEnvVars.API_BASE_URL,
    ApifyEnvVars.API_PUBLIC_BASE_URL,
    ApifyEnvVars.CHROME_EXECUTABLE_PATH,
    ApifyEnvVars.CONTAINER_URL,
    ApifyEnvVars.DEFAULT_BROWSER_PATH,
    ApifyEnvVars.DEFAULT_DATASET_ID,
    ApifyEnvVars.DEFAULT_KEY_VALUE_STORE_ID,
    ApifyEnvVars.DEFAULT_REQUEST_QUEUE_ID,
    ApifyEnvVars.FACT,
    ApifyEnvVars.INPUT_KEY,
    ApifyEnvVars.INPUT_SECRETS_PRIVATE_KEY_FILE,
    ApifyEnvVars.INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE,
    ApifyEnvVars.LOCAL_STORAGE_DIR,
    ApifyEnvVars.LOG_FORMAT,
    ApifyEnvVars.META_ORIGIN,
    ApifyEnvVars.PROXY_HOSTNAME,
    ApifyEnvVars.PROXY_PASSWORD,
    ApifyEnvVars.PROXY_STATUS_URL,
    ApifyEnvVars.SDK_LATEST_VERSION,
    ApifyEnvVars.TOKEN,
    ApifyEnvVars.USER_ID,
    ApifyEnvVars.WORKFLOW_KEY,
]

STRING_ENV_VARS: List[_STRING_ENV_VARS_TYPE] = list(get_args(_STRING_ENV_VARS_TYPE))


class ActorExitCodes(int, Enum):
    """Usual actor exit codes."""

    SUCCESS = 0
    ERROR_USER_FUNCTION_THREW = 91


class StorageTypes(str, Enum):
    """Possible Apify storage types."""

    DATASET = 'Dataset'
    KEY_VALUE_STORE = 'Key-value store'
    REQUEST_QUEUE = 'Request queue'


DEFAULT_API_PARAM_LIMIT = 1000

REQUEST_ID_LENGTH = 15

REQUEST_QUEUE_HEAD_MAX_LIMIT = 1000

EVENT_LISTENERS_TIMEOUT_SECS = 5



import io
import json
import mimetypes
import os
import pathlib
import uuid
import warnings
from datetime import datetime
from operator import itemgetter
from typing import TYPE_CHECKING, Any, AsyncIterator, Dict, Optional, Union

import aioshutil

from ..._utils import (
    _force_rename,
    _guess_file_extension,
    _is_file_or_bytes,
    _is_uuid,
    _json_dumps,
    _maybe_parse_body,
    _raise_on_duplicate_storage,
    _raise_on_non_existing_storage,
)
from ...consts import DEFAULT_API_PARAM_LIMIT, StorageTypes
from ..file_storage_utils import _set_or_delete_key_value_store_record, _update_metadata

if TYPE_CHECKING:
    from ..memory_storage import MemoryStorage

DEFAULT_LOCAL_FILE_EXTENSION = 'bin'


class KeyValueStoreClient:
    """TODO: docs."""

    _id: str
    _key_value_store_directory: str
    _client: 'MemoryStorage'
    _name: Optional[str]
    _key_value_entries: Dict[str, Dict]
    _created_at: datetime
    _accessed_at: datetime
    _modified_at: datetime

    def __init__(self, *, base_storage_directory: str, client: 'MemoryStorage', id: Optional[str] = None, name: Optional[str] = None) -> None:
        """TODO: docs."""
        self._id = str(uuid.uuid4()) if id is None else id
        self._key_value_store_directory = os.path.join(base_storage_directory, name or self._id)
        self._client = client
        self._name = name
        self._key_value_entries = {}
        self._created_at = datetime.utcnow()
        self._accessed_at = datetime.utcnow()
        self._modified_at = datetime.utcnow()

    async def get(self) -> Optional[Dict]:
        """TODO: docs."""
        found = _find_or_cache_key_value_store_by_possible_id(client=self._client, entry_name_or_id=self._name or self._id)

        if found:
            await found._update_timestamps(False)
            return found.to_key_value_store_info()

        return None

    async def update(self, *, name: Optional[str] = None) -> Dict:
        """TODO: docs."""
        # Check by id
        existing_store_by_id = _find_or_cache_key_value_store_by_possible_id(client=self._client, entry_name_or_id=self._name or self._id)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self._id)

        # Skip if no changes
        if name is None:
            return existing_store_by_id.to_key_value_store_info()

        # Check that name is not in use already
        existing_store_by_name = next(
            (store for store in self._client._key_value_stores_handled if store._name and store._name.lower() == name.lower()), None)

        if existing_store_by_name is not None:
            _raise_on_duplicate_storage(StorageTypes.KEY_VALUE_STORE, 'name', name)

        existing_store_by_id._name = name

        previous_dir = existing_store_by_id._key_value_store_directory

        existing_store_by_id._key_value_store_directory = os.path.join(self._client._key_value_stores_directory, name)

        await _force_rename(previous_dir, existing_store_by_id._key_value_store_directory)

        # Update timestamps
        await existing_store_by_id._update_timestamps(True)

        return existing_store_by_id.to_key_value_store_info()

    async def delete(self) -> None:
        """TODO: docs."""
        store = next((store for store in self._client._key_value_stores_handled if store._id == self._id), None)

        if store is not None:
            self._client._key_value_stores_handled.remove(store)
            store._key_value_entries.clear()

            if os.path.exists(store._key_value_store_directory):
                await aioshutil.rmtree(store._key_value_store_directory)

    async def list_keys(self, *, limit: int = DEFAULT_API_PARAM_LIMIT, exclusive_start_key: Optional[str] = None) -> Dict:
        """TODO: docs."""
        # Check by id
        existing_store_by_id = _find_or_cache_key_value_store_by_possible_id(self._client, self._name or self._id)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self._id)

        items = []

        for record in existing_store_by_id._key_value_entries.values():
            size = len(record['value'])
            items.append({
                'key': record['key'],
                'size': size,
            })

        if len(items) == 0:
            return {
                'count': len(items),
                'limit': limit,
                'exclusiveStartKey': exclusive_start_key,
                'isTruncated': False,
                'nextExclusiveStartKey': None,
                'items': items,
            }

        # Lexically sort to emulate the API
        items = sorted(items, key=itemgetter('key'))

        truncated_items = items
        if exclusive_start_key is not None:
            key_pos = next((idx for idx, i in enumerate(items) if i['key'] == exclusive_start_key), None)
            if key_pos is not None:
                truncated_items = items[key_pos + 1:]

        limited_items = truncated_items[:limit]

        last_item_in_store = items[-1]
        last_selected_item = limited_items[-1]
        is_last_selected_item_absolutely_last = last_item_in_store == last_selected_item
        next_exclusive_start_key = None if is_last_selected_item_absolutely_last else last_selected_item['key']

        await existing_store_by_id._update_timestamps(False)

        return {
            'count': len(items),
            'limit': limit,
            'exclusiveStartKey': exclusive_start_key,
            'isTruncated': not is_last_selected_item_absolutely_last,
            'nextExclusiveStartKey': next_exclusive_start_key,
            'items': limited_items,
        }

    async def _get_record_internal(self, key: str, as_bytes: bool = False) -> Optional[Dict]:
        # Check by id
        existing_store_by_id = _find_or_cache_key_value_store_by_possible_id(self._client, self._name or self._id)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self._id)

        entry = existing_store_by_id._key_value_entries.get(key)

        if entry is None:
            return None

        record = {
            'key': entry['key'],
            'value': entry['value'],
            # To guess the type, we need a real file name, not just the extension. e.g. 'file.json' instead of 'json'
            'contentType': entry.get('content_type') or mimetypes.guess_type(f"file.{entry['extension']}")[0],
        }

        if not as_bytes:
            record['value'] = _maybe_parse_body(record['value'], record['contentType'])

        await existing_store_by_id._update_timestamps(False)

        return record

    async def get_record(self, key: str) -> Optional[Dict]:
        """TODO: docs."""
        return await self._get_record_internal(key)

    async def get_record_as_bytes(self, key: str) -> Optional[Dict]:
        """TODO: docs."""
        return await self._get_record_internal(key, as_bytes=True)

    async def stream_record(self, _key: str) -> AsyncIterator[Optional[Dict]]:
        """TODO: docs."""
        raise NotImplementedError('This method is not supported in local memory storage')

    async def set_record(self, key: str, value: Any, content_type: Optional[str] = None) -> None:
        """TODO: docs."""
        # Check by id
        existing_store_by_id = _find_or_cache_key_value_store_by_possible_id(self._client, self._name or self._id)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self._id)

        if isinstance(value, io.IOBase):
            raise NotImplementedError('File-like values are not supported in local memory storage')

        if content_type is None:
            if _is_file_or_bytes(value):
                content_type = 'application/octet-stream'
            elif isinstance(value, str):
                content_type = 'text/plain; charset=utf-8'
            else:
                content_type = 'application/json; charset=utf-8'

        extension = _guess_file_extension(content_type or '') or DEFAULT_LOCAL_FILE_EXTENSION

        if 'application/json' in content_type and not _is_file_or_bytes(value) and not isinstance(value, str):
            value = _json_dumps(value).encode('utf-8')

        record = {
            'extension': extension,
            'key': key,
            'value': value,
            'content_type': content_type,
        }

        existing_store_by_id._key_value_entries[key] = record

        await existing_store_by_id._update_timestamps(True)
        await _set_or_delete_key_value_store_record(
            entity_directory=existing_store_by_id._key_value_store_directory,
            persist_storage=self._client._persist_storage,
            record=record,
            should_set=True,
            write_metadata=self._client._write_metadata,
        )

    async def delete_record(self, key: str) -> None:
        """TODO: docs."""
        # Check by id
        existing_store_by_id = _find_or_cache_key_value_store_by_possible_id(self._client, self._name or self._id)

        if existing_store_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self._id)

        entry = existing_store_by_id._key_value_entries.get(key)

        if entry is not None:
            del existing_store_by_id._key_value_entries[key]
            await existing_store_by_id._update_timestamps(True)
            await _set_or_delete_key_value_store_record(
                entity_directory=existing_store_by_id._key_value_store_directory,
                persist_storage=self._client._persist_storage,
                record=entry,
                should_set=False,
                write_metadata=self._client._write_metadata,
            )

    def to_key_value_store_info(self) -> Dict:
        """TODO: docs."""
        return {
            'id': self._id,
            'name': self._name,
            'accessedAt': self._accessed_at,
            'createdAt': self._created_at,
            'modifiedAt': self._modified_at,
            'userId': '1',
        }

    async def _update_timestamps(self, has_been_modified: bool) -> None:
        """TODO: docs."""
        self._accessed_at = datetime.utcnow()

        if has_been_modified:
            self._modified_at = datetime.utcnow()

        kv_store_info = self.to_key_value_store_info()
        await _update_metadata(data=kv_store_info, entity_directory=self._key_value_store_directory, write_metadata=self._client._write_metadata)


def _find_or_cache_key_value_store_by_possible_id(client: 'MemoryStorage', entry_name_or_id: str) -> Optional['KeyValueStoreClient']:
    # First check memory cache
    found = next((store for store in client._key_value_stores_handled
                  if store._id == entry_name_or_id or (store._name and store._name.lower() == entry_name_or_id.lower())), None)

    if found is not None:
        return found

    key_value_store_dir = os.path.join(client._key_value_stores_directory, entry_name_or_id)
    # Check if directory exists
    if not os.access(key_value_store_dir, os.F_OK):
        return None

    id: Union[str, None] = None
    name: Union[str, None] = None
    created_at = datetime.utcnow()
    accessed_at = datetime.utcnow()
    modified_at = datetime.utcnow()
    internal_records: Dict[str, Dict] = {}

    # Access the key value store folder
    for entry in os.scandir(key_value_store_dir):
        if entry.is_file():
            if entry.name == '__metadata__.json':
                # We have found the store metadata file, build out information based on it
                with open(os.path.join(key_value_store_dir, entry.name), encoding='utf8') as f:
                    metadata = json.load(f)
                id = metadata['id']
                name = metadata['name']
                created_at = datetime.strptime(metadata['createdAt'], '%Y-%m-%dT%H:%M:%S.%fZ')
                accessed_at = datetime.strptime(metadata['accessedAt'], '%Y-%m-%dT%H:%M:%S.%fZ')
                modified_at = datetime.strptime(metadata['modifiedAt'], '%Y-%m-%dT%H:%M:%S.%fZ')

                continue

            if '.__metadata__.' in entry.name:
                # This is an entry's metadata file, we can use it to create/extend the record
                with open(os.path.join(key_value_store_dir, entry.name), encoding='utf8') as f:
                    metadata = json.load(f)

                new_record = {
                    **internal_records.get(metadata['key'], {}),
                    **metadata,
                }

                internal_records[metadata['key']] = new_record

                continue

            with open(os.path.join(key_value_store_dir, entry.name), 'rb') as f:
                file_content = f.read()
            file_extension = pathlib.Path(entry.name).suffix
            content_type, _ = mimetypes.guess_type(entry.name)
            if content_type is None:
                content_type = 'text/plain'
            extension = _guess_file_extension(content_type)

            if file_extension == '':
                # We need to override and then restore the warnings filter so that the warning gets printed out,
                # Otherwise it would be silently swallowed
                with warnings.catch_warnings():
                    warnings.simplefilter('always')
                    warnings.warn(
                        f"""Key-value entry "{entry.name}" for store {entry_name_or_id} does not have a file extension, assuming it as text.
                        If you want to have correct interpretation of the file, you should add a file extension to the entry.""",
                        Warning,
                        stacklevel=2,
                    )
            elif 'application/json' in content_type:
                try:
                    # Try parsing the JSON ahead of time (not ideal but solves invalid files being loaded into stores)
                    json.loads(file_content)
                except json.JSONDecodeError:
                    # We need to override and then restore the warnings filter so that the warning gets printed out,
                    # Otherwise it would be silently swallowed
                    with warnings.catch_warnings():
                        warnings.simplefilter('always')
                        warnings.warn(
                            (f'Key-value entry "{entry.name}" for store {entry_name_or_id} has invalid JSON content'
                             'and will be ignored from the store.'),
                            Warning,
                            stacklevel=2,
                        )
                    continue

            name_split = entry.name.split('.')

            if file_extension != '':
                name_split.pop()

            key = '.'.join(name_split)

            new_record = {
                'key': key,
                'extension': extension,
                'value': file_content,
                'content_type': content_type,
                **internal_records.get(key, {}),
            }

            internal_records[key] = new_record

    if id is None and name is None:
        is_uuid = _is_uuid(entry_name_or_id)

        if is_uuid:
            id = entry_name_or_id
        else:
            name = entry_name_or_id

    new_client = KeyValueStoreClient(base_storage_directory=client._key_value_stores_directory, client=client, id=id, name=name)

    # Overwrite properties
    new_client._accessed_at = accessed_at
    new_client._created_at = created_at
    new_client._modified_at = modified_at

    for key, record in internal_records.items():
        new_client._key_value_entries[key] = record

    client._key_value_stores_handled.append(new_client)

    return new_client



from typing import Optional

from ._utils import _fetch_and_parse_env_var
from .consts import ApifyEnvVars


class Configuration:
    """A class for specifying the configuration of an actor.

    Can be used either globally via `Configuration.get_global_configuration()`,
    or it can be specific to each `Actor` instance on the `actor.config` property.
    """

    _default_instance: Optional['Configuration'] = None

    def __init__(self) -> None:
        """Create a `Configuration` instance."""
        self.actor_build_id = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_BUILD_ID)
        self.actor_build_number = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_BUILD_NUMBER)
        self.actor_events_ws_url = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_EVENTS_WS_URL)
        self.actor_id = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_ID)
        self.actor_run_id = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_RUN_ID)
        self.actor_task_id = _fetch_and_parse_env_var(ApifyEnvVars.ACTOR_TASK_ID)
        self.api_base_url = _fetch_and_parse_env_var(ApifyEnvVars.API_BASE_URL, 'https://api.apify.com')
        self.api_public_base_url = _fetch_and_parse_env_var(ApifyEnvVars.API_PUBLIC_BASE_URL, 'https://api.apify.com')
        self.chrome_executable_path = _fetch_and_parse_env_var(ApifyEnvVars.CHROME_EXECUTABLE_PATH)
        self.container_port = _fetch_and_parse_env_var(ApifyEnvVars.CONTAINER_PORT, 4321)
        self.container_url = _fetch_and_parse_env_var(ApifyEnvVars.CONTAINER_URL, 'http://localhost:4321')
        self.dedicated_cpus = _fetch_and_parse_env_var(ApifyEnvVars.DEDICATED_CPUS)
        self.default_browser_path = _fetch_and_parse_env_var(ApifyEnvVars.DEFAULT_BROWSER_PATH)
        self.default_dataset_id = _fetch_and_parse_env_var(ApifyEnvVars.DEFAULT_DATASET_ID, 'default')
        self.default_key_value_store_id = _fetch_and_parse_env_var(ApifyEnvVars.DEFAULT_KEY_VALUE_STORE_ID, 'default')
        self.default_request_queue_id = _fetch_and_parse_env_var(ApifyEnvVars.DEFAULT_REQUEST_QUEUE_ID, 'default')
        self.disable_browser_sandbox = _fetch_and_parse_env_var(ApifyEnvVars.DISABLE_BROWSER_SANDBOX, False)
        self.headless = _fetch_and_parse_env_var(ApifyEnvVars.HEADLESS, True)
        self.input_key = _fetch_and_parse_env_var(ApifyEnvVars.INPUT_KEY, 'INPUT')
        self.input_secrets_private_key_file = _fetch_and_parse_env_var(ApifyEnvVars.INPUT_SECRETS_PRIVATE_KEY_FILE)
        self.input_secrets_private_key_passphrase = _fetch_and_parse_env_var(ApifyEnvVars.INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE)
        self.is_at_home = _fetch_and_parse_env_var(ApifyEnvVars.IS_AT_HOME, False)
        self.memory_mbytes = _fetch_and_parse_env_var(ApifyEnvVars.MEMORY_MBYTES)
        self.meta_origin = _fetch_and_parse_env_var(ApifyEnvVars.META_ORIGIN)
        self.metamorph_after_sleep_millis = _fetch_and_parse_env_var(ApifyEnvVars.METAMORPH_AFTER_SLEEP_MILLIS, 300000)
        self.persist_state_interval_millis = _fetch_and_parse_env_var(ApifyEnvVars.PERSIST_STATE_INTERVAL_MILLIS, 60000)
        self.persist_storage = _fetch_and_parse_env_var(ApifyEnvVars.PERSIST_STORAGE)
        self.proxy_hostname = _fetch_and_parse_env_var(ApifyEnvVars.PROXY_HOSTNAME, 'proxy.apify.com')
        self.proxy_password = _fetch_and_parse_env_var(ApifyEnvVars.PROXY_PASSWORD)
        self.proxy_port = _fetch_and_parse_env_var(ApifyEnvVars.PROXY_PORT, 8000)
        self.proxy_status_url = _fetch_and_parse_env_var(ApifyEnvVars.PROXY_STATUS_URL, 'http://proxy.apify.com')
        self.purge_on_start = _fetch_and_parse_env_var(ApifyEnvVars.PURGE_ON_START, True)
        self.started_at = _fetch_and_parse_env_var(ApifyEnvVars.STARTED_AT)
        self.timeout_at = _fetch_and_parse_env_var(ApifyEnvVars.TIMEOUT_AT)
        self.token = _fetch_and_parse_env_var(ApifyEnvVars.TOKEN)
        self.user_id = _fetch_and_parse_env_var(ApifyEnvVars.USER_ID)
        self.xvfb = _fetch_and_parse_env_var(ApifyEnvVars.XVFB, False)

        self.system_info_interval_millis = 60000
        self.max_used_cpu_ratio = 0.95

    @classmethod
    def _get_default_instance(cls) -> 'Configuration':
        if cls._default_instance is None:
            cls._default_instance = cls()

        return cls._default_instance

    @classmethod
    def get_global_configuration(cls) -> 'Configuration':
        """Retrive the global configuration.

        The global configuration applies when you call actor methods via their static versions, e.g. `Actor.init()`.
        Also accessible via `Actor.config`.
        """
        return cls._get_default_instance()



from .dataset import DatasetClient
from .dataset_collection import DatasetCollectionClient
from .key_value_store import KeyValueStoreClient
from .key_value_store_collection import KeyValueStoreCollectionClient
from .request_queue import RequestQueueClient
from .request_queue_collection import RequestQueueCollectionClient

__all__ = ['DatasetClient', 'DatasetCollectionClient', 'KeyValueStoreClient',
           'KeyValueStoreCollectionClient', 'RequestQueueClient', 'RequestQueueCollectionClient']



import os
from pathlib import Path
from typing import List, Optional

import aioshutil
from aiofiles import ospath
from aiofiles.os import rename, scandir

from .resource_clients.dataset import DatasetClient
from .resource_clients.dataset_collection import DatasetCollectionClient
from .resource_clients.key_value_store import KeyValueStoreClient
from .resource_clients.key_value_store_collection import KeyValueStoreCollectionClient
from .resource_clients.request_queue import RequestQueueClient
from .resource_clients.request_queue_collection import RequestQueueCollectionClient

"""
Memory storage emulates data storages that are available on the Apify platform.
Specifically, it emulates clients for datasets, key-value stores and request queues.
The data are held in-memory and persisted locally if `persist_storage` is True.
The metadata of the storages is also persisted if `write_metadata` is True.
"""


class MemoryStorage:
    """Class representing an in-memory storage."""

    _local_data_directory: str
    _datasets_directory: str
    _key_value_stores_directory: str
    _request_queues_directory: str
    _write_metadata: bool
    _persist_storage: bool
    _datasets_handled: List[DatasetClient]
    _key_value_stores_handled: List[KeyValueStoreClient]
    _request_queues_handled: List[RequestQueueClient]

    _purged: bool = False
    """Indicates whether a purge was already performed on this instance"""

    def __init__(
        self, *, local_data_directory: str = './storage', write_metadata: Optional[bool] = None, persist_storage: Optional[bool] = None,
    ) -> None:
        """Initialize the MemoryStorage.

        Args:
            local_data_directory (str, optional): A local directory where all data will be persisted
            persist_storage (bool, optional): Whether to persist the data to the `local_data_directory` or just keep them in memory
            write_metadata (bool, optional): Whether to persist metadata of the storages as well
        """
        self._local_data_directory = local_data_directory
        self._datasets_directory = os.path.join(self._local_data_directory, 'datasets')
        self._key_value_stores_directory = os.path.join(self._local_data_directory, 'key_value_stores')
        self._request_queues_directory = os.path.join(self._local_data_directory, 'request_queues')
        self._write_metadata = write_metadata if write_metadata is not None else '*' in os.getenv('DEBUG', '')
        self._persist_storage = persist_storage if persist_storage is not None else not any(
            os.getenv('APIFY_PERSIST_STORAGE', 'true') == s for s in ['false', '0', ''])
        self._datasets_handled = []
        self._key_value_stores_handled = []
        self._request_queues_handled = []

    def datasets(self) -> DatasetCollectionClient:
        """Retrieve the sub-client for manipulating datasets."""
        return DatasetCollectionClient(base_storage_directory=self._datasets_directory, client=self)

    def dataset(self, dataset_id: str) -> DatasetClient:
        """Retrieve the sub-client for manipulating a single dataset.

        Args:
            dataset_id (str): ID of the dataset to be manipulated
        """
        return DatasetClient(base_storage_directory=self._datasets_directory, client=self, id=dataset_id)

    def key_value_stores(self) -> KeyValueStoreCollectionClient:
        """Retrieve the sub-client for manipulating key-value stores."""
        return KeyValueStoreCollectionClient(base_storage_directory=self._key_value_stores_directory, client=self)

    def key_value_store(self, key_value_store_id: str) -> KeyValueStoreClient:
        """Retrieve the sub-client for manipulating a single key-value store.

        Args:
            key_value_store_id (str): ID of the key-value store to be manipulated
        """
        return KeyValueStoreClient(base_storage_directory=self._key_value_stores_directory, client=self, id=key_value_store_id)

    def request_queues(self) -> RequestQueueCollectionClient:
        """Retrieve the sub-client for manipulating request queues."""
        return RequestQueueCollectionClient(base_storage_directory=self._request_queues_directory, client=self)

    def request_queue(self, request_queue_id: str, *, client_key: Optional[str] = None) -> RequestQueueClient:  # noqa: U100
        """Retrieve the sub-client for manipulating a single request queue.

        Args:
            request_queue_id (str): ID of the request queue to be manipulated
            client_key (str): A unique identifier of the client accessing the request queue
        """
        return RequestQueueClient(base_storage_directory=self._request_queues_directory, client=self, id=request_queue_id)

    async def purge(self) -> None:
        """Clean up the default storage directories before the run starts.

        Specifically, `purge` cleans up:
         - local directory containing the default dataset;
         - all records from the default key-value store in the local directory, except for the "INPUT" key;
         - local directory containing the default request queue.
        """
        # Key-value stores
        if await ospath.exists(self._key_value_stores_directory):
            key_value_store_folders = await scandir(self._key_value_stores_directory)
            for key_value_store_folder in key_value_store_folders:
                if key_value_store_folder.name.startswith('__APIFY_TEMPORARY') or key_value_store_folder.name.startswith('__OLD'):
                    await self._batch_remove_files(key_value_store_folder.path)
                elif key_value_store_folder.name == 'default':
                    await self._handle_default_key_value_store(key_value_store_folder.path)

        # Datasets
        if await ospath.exists(self._datasets_directory):
            dataset_folders = await scandir(self._datasets_directory)
            for dataset_folder in dataset_folders:
                if dataset_folder.name == 'default' or dataset_folder.name.startswith('__APIFY_TEMPORARY'):
                    await self._batch_remove_files(dataset_folder.path)
        # Request queues
        if await ospath.exists(self._request_queues_directory):
            request_queue_folders = await scandir(self._request_queues_directory)
            for request_queue_folder in request_queue_folders:
                if request_queue_folder.name == 'default' or request_queue_folder.name.startswith('__APIFY_TEMPORARY'):
                    await self._batch_remove_files(request_queue_folder.path)

    def teardown(self) -> None:
        """TODO: docs."""
        # We don't need to wait for anything here since we don't have worker threads for fs operations
        pass

    async def _handle_default_key_value_store(self, folder: str) -> None:
        """Remove everything from the default key-value store folder except `possible_input_keys`."""
        folder_exists = await ospath.exists(folder)
        temporary_path = os.path.normpath(os.path.join(folder, '../__APIFY_MIGRATING_KEY_VALUE_STORE__'))

        # For optimization, we want to only attempt to copy a few files from the default key-value store
        possible_input_keys = [
            'INPUT',
            'INPUT.json',
            'INPUT.bin',
            'INPUT.txt',
        ]

        if folder_exists:
            # Create a temporary folder to save important files in
            Path(temporary_path).mkdir(parents=True, exist_ok=True)

            # Go through each file and save the ones that are important
            for entity in possible_input_keys:
                original_file_path = os.path.join(folder, entity)
                temp_file_path = os.path.join(temporary_path, entity)
                try:
                    await rename(original_file_path, temp_file_path)
                except Exception:
                    # Ignore
                    pass

            # Remove the original folder and all its content
            counter = 0
            temp_path_for_old_folder = os.path.normpath(os.path.join(folder, f'../__OLD_DEFAULT_{counter}__'))
            done = False
            while not done:
                try:
                    await rename(folder, temp_path_for_old_folder)
                    done = True
                except Exception:
                    counter += 1
                    temp_path_for_old_folder = os.path.normpath(os.path.join(folder, f'../__OLD_DEFAULT_{counter}__'))

            # Replace the temporary folder with the original folder
            await rename(temporary_path, folder)

            # Remove the old folder
            await self._batch_remove_files(temp_path_for_old_folder)

    async def _batch_remove_files(self, folder: str, counter: int = 0) -> None:
        folder_exists = await ospath.exists(folder)

        if folder_exists:
            # TODO: the startswith condition is always False, it's also broken in crawlee...
            temporary_folder = folder if folder.startswith('__APIFY_TEMPORARY_') else os.path.normpath(
                os.path.join(folder, f'../__APIFY_TEMPORARY_{counter}__'))

            try:
                # Rename the old folder to the new one to allow background deletions
                await rename(folder, temporary_folder)
            except Exception:
                # Folder exists already, try again with an incremented counter
                return await self._batch_remove_files(folder, counter + 1)

            await aioshutil.rmtree(temporary_folder, ignore_errors=True)



from typing import Any, AsyncIterator, Dict, Optional, Tuple, TypeVar, Union, overload

from apify_client import ApifyClientAsync
from apify_client.clients import KeyValueStoreClientAsync

from .._utils import _wrap_internal
from ..config import Configuration
from ..memory_storage import MemoryStorage
from ..memory_storage.resource_clients import KeyValueStoreClient
from .storage_manager import StorageManager

T = TypeVar('T')


class KeyValueStore:
    """TODO: docs."""

    _id: str
    _name: Optional[str]
    _client: Union[KeyValueStoreClientAsync, KeyValueStoreClient]
    _config: Configuration

    def __init__(self, id: str, name: Optional[str], client: Union[ApifyClientAsync, MemoryStorage]) -> None:
        """TODO: docs (constructor should be "internal")."""
        self.get_value = _wrap_internal(self._get_value_internal, self.get_value)  # type: ignore
        self.set_value = _wrap_internal(self._set_value_internal, self.set_value)  # type: ignore
        self._id = id
        self._name = name
        self._client = client.key_value_store(self._id)
        self._config = Configuration.get_global_configuration()  # We always use the global config

    @classmethod
    async def _create_instance(cls, store_id_or_name: str, client: Union[ApifyClientAsync, MemoryStorage]) -> 'KeyValueStore':
        """TODO: docs."""
        key_value_store_client = client.key_value_store(store_id_or_name)
        key_value_store_info = await key_value_store_client.get()
        if not key_value_store_info:
            key_value_store_info = await client.key_value_stores().get_or_create(name=store_id_or_name)

        return KeyValueStore(key_value_store_info['id'], key_value_store_info.get('name'), client)

    @classmethod
    def _get_default_name(cls, config: Configuration) -> str:
        return config.default_key_value_store_id

    @overload
    @classmethod
    async def get_value(cls, key: str) -> Any:  # noqa: U100
        ...

    @overload
    @classmethod
    async def get_value(cls, key: str, default_value: T) -> T:  # noqa: U100
        ...

    @overload
    @classmethod
    async def get_value(cls, key: str, default_value: Optional[T] = None) -> Optional[T]:  # noqa: U100
        ...

    @classmethod
    async def get_value(cls, key: str, default_value: Optional[T] = None) -> Optional[T]:
        """TODO: docs."""
        store = await cls.open()
        return await store.get_value(key, default_value)

    async def _get_value_internal(self, key: str, default_value: Optional[T] = None) -> Optional[T]:
        record = await self._client.get_record(key)
        return record['value'] if record else default_value

    async def for_each_key(self, exclusive_start_key: Optional[str] = None) -> AsyncIterator[Tuple[Dict, int, int]]:
        """TODO: docs."""
        index = 0
        while True:
            list_keys = await self._client.list_keys(exclusive_start_key=exclusive_start_key)
            for item in list_keys['items']:
                yield item, index, item['size']
                index += 1

            if not list_keys['isTruncated']:
                break
            exclusive_start_key = list_keys['nextExclusiveStartKey']

    @classmethod
    async def set_value(cls, key: str, value: Optional[T], content_type: Optional[str] = None) -> None:
        """TODO: docs."""
        store = await cls.open()
        return await store.set_value(key, value, content_type)

    async def _set_value_internal(self, key: str, value: Optional[T], content_type: Optional[str] = None) -> None:
        if value is None:
            return await self._client.delete_record(key)

        # TODO: Why does crawlee run another stringify here?
        # value = maybeStringify(value, optionsCopy)

        return await self._client.set_record(key, value, content_type)

    async def drop(self) -> None:
        """TODO: docs."""
        await self._client.delete()
        await StorageManager.close_storage(self.__class__, self._id, self._name)

    @classmethod
    async def open(cls, store_id_or_name: Optional[str] = None, config: Optional[Configuration] = None) -> 'KeyValueStore':
        """TODO: docs."""
        return await StorageManager.open_storage(cls, store_id_or_name, None, config)



from typing import Optional, Union

from apify_client import ApifyClientAsync
from apify_client.clients import RequestQueueClientAsync

from ..config import Configuration
from ..memory_storage import MemoryStorage
from ..memory_storage.resource_clients import RequestQueueClient
from ._utils import _crypto_random_object_id


class RequestQueue:
    """TODO: docs."""

    _id: str
    _name: Optional[str]
    _client: Union[RequestQueueClientAsync, RequestQueueClient]
    _config: Configuration
    _client_key = _crypto_random_object_id()

    def __init__(self, id: str, name: Optional[str], client: Union[ApifyClientAsync, MemoryStorage]) -> None:
        """TODO: docs (constructor should be "internal")."""
        self._id = id
        self._name = name
        self._client = client.request_queue(self._id, client_key=self._client_key)
        self._config = Configuration.get_global_configuration()  # We always use the global config

    @classmethod
    async def _create_instance(cls, request_queue_id_or_name: str, client: Union[ApifyClientAsync, MemoryStorage]) -> 'RequestQueue':
        request_queue_client = client.request_queue(request_queue_id_or_name)
        request_queue_info = await request_queue_client.get()
        if not request_queue_info:
            request_queue_info = await client.request_queues().get_or_create(name=request_queue_id_or_name)

        return RequestQueue(request_queue_info['id'], request_queue_info.get('name'), client)

    @classmethod
    def _get_default_name(cls, config: Configuration) -> str:
        return config.default_request_queue_id



from operator import itemgetter
from typing import TYPE_CHECKING, Dict, Optional

from apify_client._utils import ListPage

from ..file_storage_utils import _update_metadata
from .dataset import DatasetClient, _find_or_cache_dataset_by_possible_id

if TYPE_CHECKING:
    from ..memory_storage import MemoryStorage


class DatasetCollectionClient:
    """TODO: docs."""

    _datasets_directory: str
    _client: 'MemoryStorage'

    def __init__(self, *, base_storage_directory: str, client: 'MemoryStorage') -> None:
        """TODO: docs."""
        self._datasets_directory = base_storage_directory
        self._client = client

    def list(self) -> ListPage:
        """TODO: docs."""
        def map_store(store: DatasetClient) -> Dict:
            return store.to_dataset_info()
        return ListPage({
            'total': len(self._client._datasets_handled),
            'count': len(self._client._datasets_handled),
            'offset': 0,
            'limit': len(self._client._datasets_handled),
            'desc': False,
            'items': sorted(map(map_store, self._client._datasets_handled), key=itemgetter('createdAt')),
        })

    async def get_or_create(self, *, name: Optional[str] = None, _schema: Optional[Dict] = None) -> Dict:
        """TODO: docs."""
        if name:
            found = _find_or_cache_dataset_by_possible_id(client=self._client, entry_name_or_id=name)

            if found:
                return found.to_dataset_info()

        new_dataset = DatasetClient(name=name, base_storage_directory=self._datasets_directory, client=self._client)
        self._client._datasets_handled.append(new_dataset)

        dataset_info = new_dataset.to_dataset_info()

        # Write to the disk
        await _update_metadata(data=dataset_info, entity_directory=new_dataset._dataset_directory, write_metadata=self._client._write_metadata)

        return dataset_info



from typing import TYPE_CHECKING, Dict, Optional, Type, TypeVar, Union, cast

from apify_client import ApifyClientAsync

from ..config import Configuration
from ..memory_storage import MemoryStorage
from ..storage_client_manager import StorageClientManager

if TYPE_CHECKING:
    from .dataset import Dataset
    from .key_value_store import KeyValueStore
    from .request_queue import RequestQueue

T = TypeVar('T', 'Dataset', 'KeyValueStore', 'RequestQueue')


async def _purge_default_storages(client: Union[ApifyClientAsync, MemoryStorage]) -> None:
    if isinstance(client, MemoryStorage) and not client._purged:
        client._purged = True
        await client.purge()


class StorageManager:
    """TODO: docs."""

    _default_instance: Optional['StorageManager'] = None
    _cache: Dict[Type[Union['Dataset', 'KeyValueStore', 'RequestQueue']], Dict[str, Union['Dataset', 'KeyValueStore', 'RequestQueue']]]
    _config: Configuration

    def __init__(self) -> None:
        """TODO: docs."""
        self._cache = {}
        self._config = Configuration.get_global_configuration()

    @classmethod
    def _get_default_instance(cls) -> 'StorageManager':
        if cls._default_instance is None:
            cls._default_instance = cls()

        return cls._default_instance

    @classmethod
    async def open_storage(
        cls,
        storage_class: Type[T],
        store_id_or_name: Optional[str] = None,
        client: Optional[Union[ApifyClientAsync, MemoryStorage]] = None,
        config: Optional[Configuration] = None,
    ) -> T:
        """TODO: docs."""
        storage_manager = StorageManager._get_default_instance()
        used_config = config or storage_manager._config
        used_client = client or StorageClientManager.get_storage_client()

        # Create cache for the given storage class if missing
        if storage_class not in storage_manager._cache:
            storage_manager._cache[storage_class] = {}

        # Fetch default name
        if not store_id_or_name:
            store_id_or_name = storage_class._get_default_name(used_config)

        # Try to get the storage instance from cache
        storage = storage_manager._cache[storage_class].get(store_id_or_name, None)
        if storage is not None:
            # This cast is needed since we're storing all storages in one union dictionary
            return cast(T, storage)

        # Purge default storages if configured
        if used_config.purge_on_start:
            await _purge_default_storages(used_client)

        # Create the storage
        storage = await storage_class._create_instance(store_id_or_name, used_client)

        # Cache by id and name
        storage_manager._cache[storage_class][storage._id] = storage
        if storage._name is not None:
            storage_manager._cache[storage_class][storage._name] = storage
        return storage

    @classmethod
    async def close_storage(cls, storage_class: Type[Union['Dataset', 'KeyValueStore', 'RequestQueue']], id: str, name: Optional[str]) -> None:
        """TODO: docs."""
        storage_manager = StorageManager._get_default_instance()
        del storage_manager._cache[storage_class][id]
        if name is not None:
            del storage_manager._cache[storage_class][name]



from typing import Optional, Union

from apify_client import ApifyClientAsync

from .config import Configuration
from .memory_storage import MemoryStorage


class StorageClientManager:
    """TODO: docs."""

    _config: Configuration

    _client: Union[ApifyClientAsync, MemoryStorage]

    _default_instance: Optional['StorageClientManager'] = None

    def __init__(self) -> None:
        """TODO: docs."""
        self._config = Configuration.get_global_configuration()
        self._client = MemoryStorage(persist_storage=self._config.persist_storage)

    @classmethod
    def get_storage_client(cls) -> Union[ApifyClientAsync, MemoryStorage]:
        """TODO: docs."""
        return cls._get_default_instance()._client

    @classmethod
    def set_storage_client(cls, client: Union[ApifyClientAsync, MemoryStorage]) -> None:
        """TODO: docs."""
        cls._get_default_instance()._client = client

    @classmethod
    def _get_default_instance(cls) -> 'StorageClientManager':
        if cls._default_instance is None:
            cls._default_instance = cls()

        return cls._default_instance



import os
from typing import Dict, List, Tuple

import aiofiles
from aiofiles.os import makedirs

from .._utils import _force_remove, _json_dumps


async def _update_metadata(*, data: Dict, entity_directory: str, write_metadata: bool) -> None:
    # Skip writing the actual metadata file. This is done after ensuring the directory exists so we have the directory present
    if not write_metadata:
        return

    # Ensure the directory for the entity exists
    await makedirs(entity_directory, exist_ok=True)

    # Write the metadata to the file
    file_path = os.path.join(entity_directory, '__metadata__.json')
    async with aiofiles.open(file_path, mode='wb') as f:
        await f.write(_json_dumps(data).encode('utf-8'))


async def _update_dataset_items(
    *,
    data: List[Tuple[str, Dict]],
    entity_directory: str,
    persist_storage: bool,
) -> None:
    # Skip writing files to the disk if the client has the option set to false
    if not persist_storage:
        return

    # Ensure the directory for the entity exists
    await makedirs(entity_directory, exist_ok=True)

    # Save all the new items to the disk
    for idx, item in data:
        file_path = os.path.join(entity_directory, f'{idx}.json')
        async with aiofiles.open(file_path, mode='wb') as f:
            await f.write(_json_dumps(item).encode('utf-8'))


async def _set_or_delete_key_value_store_record(
    *,
    entity_directory: str,
    persist_storage: bool,
    record: Dict,
    should_set: bool,
    write_metadata: bool,
) -> None:
    # Skip writing files to the disk if the client has the option set to false
    if not persist_storage:
        return

    # Ensure the directory for the entity exists
    await makedirs(entity_directory, exist_ok=True)

    # Create files for the record
    record_path = os.path.join(entity_directory, f"""{record['key']}.{record['extension']}""")
    record_metadata_path = os.path.join(entity_directory, f"""{record['key']}.__metadata__.json""")

    await _force_remove(record_path)
    await _force_remove(record_metadata_path)

    if should_set:
        if write_metadata:
            async with aiofiles.open(record_metadata_path, mode='wb') as f:
                await f.write(_json_dumps({
                    'key': record['key'],
                    'contentType': record.get('content_type') or 'unknown/no content type',
                    'extension': record['extension'],
                }).encode('utf-8'))

        # Convert to bytes if string
        if isinstance(record['value'], str):
            record['value'] = record['value'].encode('utf-8')

        async with aiofiles.open(record_path, mode='wb') as f:
            await f.write(record['value'])


async def _update_request_queue_item(
    *,
    request_id: str,
    request: Dict,
    entity_directory: str,
    persist_storage: bool,
) -> None:
    # Skip writing files to the disk if the client has the option set to false
    if not persist_storage:
        return

    # Ensure the directory for the entity exists
    await makedirs(entity_directory, exist_ok=True)

    # Write the request to the file
    file_path = os.path.join(entity_directory, f'{request_id}.json')
    async with aiofiles.open(file_path, mode='wb') as f:
        await f.write(_json_dumps(request).encode('utf-8'))


async def _delete_request(*, request_id: str, entity_directory: str) -> None:
    # Ensure the directory for the entity exists
    await makedirs(entity_directory, exist_ok=True)

    file_path = os.path.join(entity_directory, f'{request_id}.json')
    await _force_remove(file_path)



import csv
import io
from typing import AsyncIterator, Dict, List, Optional, Union

from apify_client import ApifyClientAsync
from apify_client._utils import ListPage
from apify_client.clients import DatasetClientAsync

from .._types import JSONSerializable
from .._utils import _wrap_internal
from ..config import Configuration
from ..memory_storage import MemoryStorage
from ..memory_storage.resource_clients import DatasetClient
from .key_value_store import KeyValueStore
from .storage_manager import StorageManager


class Dataset:
    """TODO: docs."""

    _id: str
    _name: Optional[str]
    _client: Union[DatasetClientAsync, DatasetClient]
    _config: Configuration

    def __init__(self, id: str, name: Optional[str], client: Union[ApifyClientAsync, MemoryStorage]) -> None:
        """TODO: docs (constructor should be "internal")."""
        self.get_data = _wrap_internal(self._get_data_internal, self.get_data)  # type: ignore
        self.push_data = _wrap_internal(self._push_data_internal, self.push_data)  # type: ignore
        self.export_to_json = _wrap_internal(self._export_to_json_internal, self.export_to_json)  # type: ignore
        self.export_to_csv = _wrap_internal(self._export_to_csv_internal, self.export_to_csv)  # type: ignore
        self._id = id
        self._name = name
        self._client = client.dataset(self._id)
        self._config = Configuration.get_global_configuration()  # We always use the global config

    @classmethod
    async def _create_instance(cls, dataset_id_or_name: str, client: Union[ApifyClientAsync, MemoryStorage]) -> 'Dataset':
        dataset_client = client.dataset(dataset_id_or_name)
        dataset_info = await dataset_client.get()
        if not dataset_info:
            dataset_info = await client.datasets().get_or_create(name=dataset_id_or_name)

        return Dataset(dataset_info['id'], dataset_info.get('name'), client)

    @classmethod
    def _get_default_name(cls, config: Configuration) -> str:
        return config.default_dataset_id

    @classmethod
    async def push_data(cls, data: JSONSerializable) -> None:
        """TODO: docs."""
        dataset = await cls.open()
        return await dataset.push_data(data)

    async def _push_data_internal(self, data: JSONSerializable) -> None:
        # const dispatch = (payload: string) => this.client.pushItems(payload);
        # const limit = MAX_PAYLOAD_SIZE_BYTES - Math.ceil(MAX_PAYLOAD_SIZE_BYTES * SAFETY_BUFFER_PERCENT);

        # // Handle singular Objects
        # if (!Array.isArray(data)) {
        #     const payload = checkAndSerialize(data, limit);
        #     return dispatch(payload);
        # }

        # // Handle Arrays
        # const payloads = data.map((item, index) => checkAndSerialize(item, limit, index));
        # const chunks = chunkBySize(payloads, limit);

        # // Invoke client in series to preserve order of data
        # for (const chunk of chunks) {
        #     await dispatch(chunk);
        # }
        # TODO: Implement the size chunking mechanism from crawlee...
        # limit = MAX_PAYLOAD_SIZE_BYTES - math.ceil(MAX_PAYLOAD_SIZE_BYTES * SAFETY_BUFFER_PERCENT)
        await self._client.push_items(data)

    @classmethod
    async def get_data(
        cls,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        clean: Optional[bool] = None,
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,
        omit: Optional[List[str]] = None,
        unwind: Optional[str] = None,
        skip_empty: Optional[bool] = None,
        skip_hidden: Optional[bool] = None,
        flatten: Optional[List[str]] = None,
        view: Optional[str] = None,
    ) -> ListPage:
        """TODO: docs."""
        dataset = await cls.open()
        return await dataset.get_data(
            offset=offset,
            limit=limit,
            desc=desc,
            clean=clean,
            fields=fields,
            omit=omit,
            unwind=unwind,
            skip_empty=skip_empty,
            skip_hidden=skip_hidden,
            flatten=flatten,
            view=view,
        )

    async def _get_data_internal(
        self,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        clean: Optional[bool] = None,
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,
        omit: Optional[List[str]] = None,
        unwind: Optional[str] = None,
        skip_empty: Optional[bool] = None,
        skip_hidden: Optional[bool] = None,
        flatten: Optional[List[str]] = None,
        view: Optional[str] = None,
    ) -> ListPage:
        # try {
        #     return await this.client.listItems(options);
        # } catch (e) {
        #     const error = e as Error;
        #     if (error.message.includes('Cannot create a string longer than')) {
        #         throw new Error('dataset.getData(): The response is too large for parsing. You can fix this by lowering the "limit" option.');
        #     }
        #     throw e;
        # }
        # TODO: Simulate the above error in Python and handle accordingly...
        return await self._client.list_items(
            offset=offset,
            limit=limit,
            desc=desc,
            clean=clean,
            fields=fields,
            omit=omit,
            unwind=unwind,
            skip_empty=skip_empty,
            skip_hidden=skip_hidden,
            flatten=flatten,
            view=view,
        )

    async def export_to(
        self,
        key: str,
        *,
        to_key_value_store: Optional[str] = None,
        content_type: Optional[str] = None,
    ) -> None:
        """TODO: docs."""
        key_value_store = await KeyValueStore.open(to_key_value_store)
        items: List[Dict] = []
        limit = 1000
        offset = 0
        while True:
            list_items = await self._client.list_items(limit=limit, offset=offset)
            items.extend(list_items.items)
            if list_items.total <= offset + list_items.count:
                break
            offset += list_items.count

        if len(items) == 0:
            raise ValueError('Cannot export an empty dataset')

        if content_type == 'text/csv':
            output = io.StringIO()
            writer = csv.writer(output, quoting=csv.QUOTE_MINIMAL)  # TODO: Compare quoting behavior with TS impl
            writer.writerows([items[0].keys(), *[item.values() for item in items]])
            value = output.getvalue()
            return await key_value_store.set_value(key, value, content_type)

        if content_type == 'application/json':
            return await key_value_store.set_value(key, items)

        raise ValueError(f'Unsupported content type: {content_type}')

    @classmethod
    async def export_to_json(
        cls,
        key: str,
        *,
        from_dataset: Optional[str] = None,
        to_key_value_store: Optional[str] = None,
    ) -> None:
        """TODO: docs."""
        dataset = await cls.open(from_dataset)
        await dataset.export_to_json(key, to_key_value_store=to_key_value_store)

    async def _export_to_json_internal(
        self,
        key: str,
        *,
        from_dataset: Optional[str] = None,  # noqa: U100
        to_key_value_store: Optional[str] = None,
    ) -> None:
        await self.export_to(key, to_key_value_store=to_key_value_store, content_type='application/json')

    @classmethod
    async def export_to_csv(
        cls,
        key: str,
        *,
        from_dataset: Optional[str] = None,
        to_key_value_store: Optional[str] = None,
    ) -> None:
        """TODO: docs."""
        dataset = await cls.open(from_dataset)
        await dataset.export_to_csv(key, to_key_value_store=to_key_value_store)

    async def _export_to_csv_internal(
        self,
        key: str,
        *,
        from_dataset: Optional[str] = None,  # noqa: U100
        to_key_value_store: Optional[str] = None,
    ) -> None:
        await self.export_to(key, to_key_value_store=to_key_value_store, content_type='text/csv')

    async def get_info(self) -> Optional[Dict]:
        """TODO: docs."""
        return await self._client.get()

    def iterate_items(  # ~forEach in TS
        self,
        *,
        offset: int = 0,
        limit: Optional[int] = None,
        clean: Optional[bool] = None,
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,
        omit: Optional[List[str]] = None,
        unwind: Optional[str] = None,
        skip_empty: Optional[bool] = None,
        skip_hidden: Optional[bool] = None,
    ) -> AsyncIterator[Dict]:
        """TODO: docs."""
        return self._client.iterate_items(
            offset=offset,
            limit=limit,
            clean=clean,
            desc=desc,
            fields=fields,
            omit=omit,
            unwind=unwind,
            skip_empty=skip_empty,
            skip_hidden=skip_hidden,
        )

    async def drop(self) -> None:
        """TODO: docs."""
        await self._client.delete()
        await StorageManager.close_storage(self.__class__, self._id, self._name)

    @classmethod
    async def open(cls, dataset_id_or_name: Optional[str] = None, config: Optional[Configuration] = None) -> 'Dataset':
        """TODO: docs."""
        return await StorageManager.open_storage(cls, dataset_id_or_name, None, config)



from .dataset import Dataset
from .key_value_store import KeyValueStore
from .request_queue import RequestQueue
from .storage_manager import StorageManager

__all__ = ['Dataset', 'KeyValueStore', 'RequestQueue', 'StorageManager']



import json
import os
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any, AsyncIterator, Dict, List, Optional, Tuple, Union

import aioshutil

from apify_client._utils import ListPage

from ..._types import JSONSerializable
from ..._utils import _force_rename, _is_uuid, _raise_on_duplicate_storage, _raise_on_non_existing_storage
from ...consts import StorageTypes
from ..file_storage_utils import _update_dataset_items, _update_metadata

if TYPE_CHECKING:
    from ..memory_storage import MemoryStorage

"""
 This is what API returns in the x-apify-pagination-limit
 header when no limit query parameter is used.
 """
LIST_ITEMS_LIMIT = 999_999_999_999

"""
 Number of characters of the dataset item file names.
 E.g.: 000000019.json - 9 digits
"""
LOCAL_ENTRY_NAME_DIGITS = 9


class DatasetClient:
    """TODO: docs."""

    _id: str
    _dataset_directory: str
    _client: 'MemoryStorage'
    _name: Optional[str]
    _dataset_entries: Dict[str, Dict]
    _created_at: datetime
    _accessed_at: datetime
    _modified_at: datetime
    _item_count = 0

    def __init__(self, *, base_storage_directory: str, client: 'MemoryStorage', id: Optional[str] = None, name: Optional[str] = None) -> None:
        """TODO: docs."""
        self._id = str(uuid.uuid4()) if id is None else id
        self._dataset_directory = os.path.join(base_storage_directory, name or self._id)
        self._client = client
        self._name = name
        self._dataset_entries = {}
        self._created_at = datetime.utcnow()
        self._accessed_at = datetime.utcnow()
        self._modified_at = datetime.utcnow()

    async def get(self) -> Optional[Dict]:
        """TODO: docs."""
        found = _find_or_cache_dataset_by_possible_id(client=self._client, entry_name_or_id=self._name or self._id)

        if found:
            await found._update_timestamps(False)
            return found.to_dataset_info()

        return None

    async def update(self, *, name: Optional[str] = None) -> Dict:
        """TODO: docs."""
        # Check by id
        existing_dataset_by_id = _find_or_cache_dataset_by_possible_id(client=self._client, entry_name_or_id=self._name or self._id)

        if existing_dataset_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.DATASET, self._id)

        # Skip if no changes
        if name is None:
            return existing_dataset_by_id.to_dataset_info()

        # Check that name is not in use already
        existing_dataset_by_name = next(
            (dataset for dataset in self._client._datasets_handled if dataset._name and dataset._name.lower() == name.lower()), None)

        if existing_dataset_by_name is not None:
            _raise_on_duplicate_storage(StorageTypes.DATASET, 'name', name)

        existing_dataset_by_id._name = name

        previous_dir = existing_dataset_by_id._dataset_directory

        existing_dataset_by_id._dataset_directory = os.path.join(self._client._datasets_directory, name)

        await _force_rename(previous_dir, existing_dataset_by_id._dataset_directory)

        # Update timestamps
        await existing_dataset_by_id._update_timestamps(True)

        return existing_dataset_by_id.to_dataset_info()

    async def delete(self) -> None:
        """TODO: docs."""
        dataset = next((dataset for dataset in self._client._datasets_handled if dataset._id == self._id), None)

        if dataset is not None:
            self._client._datasets_handled.remove(dataset)
            dataset._item_count = 0
            dataset._dataset_entries.clear()

            if os.path.exists(dataset._dataset_directory):
                await aioshutil.rmtree(dataset._dataset_directory)

    async def list_items(
        self,
        *,
        offset: Optional[int] = 0,
        limit: Optional[int] = LIST_ITEMS_LIMIT,
        clean: Optional[bool] = None,  # noqa: U100
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,  # noqa: U100
        omit: Optional[List[str]] = None,  # noqa: U100
        unwind: Optional[str] = None,  # noqa: U100
        skip_empty: Optional[bool] = None,  # noqa: U100
        skip_hidden: Optional[bool] = None,  # noqa: U100
        flatten: Optional[List[str]] = None,  # noqa: U100
        view: Optional[str] = None,  # noqa: U100
    ) -> ListPage:
        """TODO: docs."""
        # Check by id
        existing_dataset_by_id = _find_or_cache_dataset_by_possible_id(client=self._client, entry_name_or_id=self._name or self._id)

        if existing_dataset_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.DATASET, self._id)

        start, end = existing_dataset_by_id._get_start_and_end_indexes(
            max(existing_dataset_by_id._item_count - (offset or 0) - (limit or LIST_ITEMS_LIMIT), 0) if desc else offset or 0,
            limit,
        )

        items = []

        for idx in range(start, end):
            entry_number = self._generate_local_entry_name(idx)
            items.append(existing_dataset_by_id._dataset_entries[entry_number])

        await existing_dataset_by_id._update_timestamps(False)

        if desc:
            items.reverse()

        return ListPage({
            'count': len(items),
            'desc': desc or False,
            'items': items,
            'limit': limit or LIST_ITEMS_LIMIT,
            'offset': offset or 0,
            'total': existing_dataset_by_id._item_count,
        })

    async def iterate_items(
        self,
        *,
        offset: int = 0,
        limit: Optional[int] = None,
        clean: Optional[bool] = None,  # noqa: U100
        desc: Optional[bool] = None,
        fields: Optional[List[str]] = None,  # noqa: U100
        omit: Optional[List[str]] = None,  # noqa: U100
        unwind: Optional[str] = None,  # noqa: U100
        skip_empty: Optional[bool] = None,  # noqa: U100
        skip_hidden: Optional[bool] = None,  # noqa: U100
    ) -> AsyncIterator[Dict]:
        """TODO: docs."""
        cache_size = 1000
        first_item = offset

        # If there is no limit, set last_item to None until we get the total from the first API response
        if limit is None:
            last_item = None
        else:
            last_item = offset + limit

        current_offset = first_item
        while last_item is None or current_offset < last_item:
            if last_item is None:
                current_limit = cache_size
            else:
                current_limit = min(cache_size, last_item - current_offset)

            current_items_page = await self.list_items(
                offset=current_offset,
                limit=current_limit,
                desc=desc,
            )

            current_offset += current_items_page.count
            if last_item is None or current_items_page.total < last_item:
                last_item = current_items_page.total

            for item in current_items_page.items:
                yield item

    async def get_items_as_bytes(
        self,
        *,
        _item_format: str = 'json',
        _offset: Optional[int] = None,
        _limit: Optional[int] = None,
        _desc: Optional[bool] = None,
        _clean: Optional[bool] = None,
        _bom: Optional[bool] = None,
        _delimiter: Optional[str] = None,
        _fields: Optional[List[str]] = None,
        _omit: Optional[List[str]] = None,
        _unwind: Optional[str] = None,
        _skip_empty: Optional[bool] = None,
        _skip_header_row: Optional[bool] = None,
        _skip_hidden: Optional[bool] = None,
        _xml_root: Optional[str] = None,
        _xml_row: Optional[str] = None,
        _flatten: Optional[List[str]] = None,
    ) -> bytes:
        """TODO: docs."""
        raise NotImplementedError('This method is not supported in local memory storage')

    async def stream_items(
        self,
        *,
        _item_format: str = 'json',
        _offset: Optional[int] = None,
        _limit: Optional[int] = None,
        _desc: Optional[bool] = None,
        _clean: Optional[bool] = None,
        _bom: Optional[bool] = None,
        _delimiter: Optional[str] = None,
        _fields: Optional[List[str]] = None,
        _omit: Optional[List[str]] = None,
        _unwind: Optional[str] = None,
        _skip_empty: Optional[bool] = None,
        _skip_header_row: Optional[bool] = None,
        _skip_hidden: Optional[bool] = None,
        _xml_root: Optional[str] = None,
        _xml_row: Optional[str] = None,
    ) -> AsyncIterator:
        """TODO: docs."""
        raise NotImplementedError('This method is not supported in local memory storage')

    async def push_items(self, items: JSONSerializable) -> None:
        """TODO: docs."""
        # Check by id
        existing_dataset_by_id = _find_or_cache_dataset_by_possible_id(client=self._client, entry_name_or_id=self._name or self._id)

        if existing_dataset_by_id is None:
            _raise_on_non_existing_storage(StorageTypes.DATASET, self._id)

        normalized = self._normalize_items(items)

        added_ids: List[str] = []
        for entry in normalized:
            existing_dataset_by_id._item_count += 1
            idx = self._generate_local_entry_name(existing_dataset_by_id._item_count)

            existing_dataset_by_id._dataset_entries[idx] = entry
            added_ids.append(idx)

        data_entries: List[Tuple[str, Dict]] = []
        for id in added_ids:
            data_entries.append((id, existing_dataset_by_id._dataset_entries[id]))

        await existing_dataset_by_id._update_timestamps(True)

        await _update_dataset_items(
            data=data_entries,
            entity_directory=existing_dataset_by_id._dataset_directory,
            persist_storage=self._client._persist_storage,
        )

    def to_dataset_info(self) -> Dict:
        """TODO: docs."""
        return {
            'id': self._id,
            'name': self._name,
            'itemCount': self._item_count,
            'accessedAt': self._accessed_at,
            'createdAt': self._created_at,
            'modifiedAt': self._modified_at,
        }

    async def _update_timestamps(self, has_been_modified: bool) -> None:
        """TODO: docs."""
        self._accessed_at = datetime.utcnow()

        if has_been_modified:
            self._modified_at = datetime.utcnow()

        dataset_info = self.to_dataset_info()
        await _update_metadata(data=dataset_info, entity_directory=self._dataset_directory, write_metadata=self._client._write_metadata)

    def _get_start_and_end_indexes(self, offset: int, limit: Optional[int] = None) -> Tuple[int, int]:
        actual_limit = limit or self._item_count
        start = offset + 1
        end = min(offset + actual_limit, self._item_count) + 1
        return (start, end)

    def _generate_local_entry_name(self, idx: int) -> str:
        return str(idx).zfill(LOCAL_ENTRY_NAME_DIGITS)

    def _normalize_items(self, items: JSONSerializable) -> List[Dict]:
        def normalize_item(item: Any) -> Optional[Dict]:
            if type(item) is str:
                item = json.loads(item)

            if type(item) is list:
                received = ',\n'.join(item)
                raise ValueError(f'Each dataset item can only be a single JSON object, not an array. Received: [{received}]')

            if type(item) is not dict and item is not None:
                raise ValueError(f'Each dataset item must be a JSON object. Received: {item}')

            return item

        if type(items) is str:
            items = json.loads(items)

        result = list(map(normalize_item, items)) if type(items) is list else [normalize_item(items)]
        # filter(None, ..) returns items that are True
        return list(filter(None, result))


def _find_or_cache_dataset_by_possible_id(client: 'MemoryStorage', entry_name_or_id: str) -> Optional['DatasetClient']:
    # First check memory cache
    found = next((dataset for dataset in client._datasets_handled
                  if dataset._id == entry_name_or_id or (dataset._name and dataset._name.lower() == entry_name_or_id.lower())), None)

    if found is not None:
        return found

    datasets_dir = os.path.join(client._datasets_directory, entry_name_or_id)
    # Check if directory exists
    if not os.access(datasets_dir, os.F_OK):
        return None

    id: Union[str, None] = None
    name: Union[str, None] = None
    item_count = 0
    created_at = datetime.utcnow()
    accessed_at = datetime.utcnow()
    modified_at = datetime.utcnow()
    entries: Dict[str, Dict] = {}

    has_seen_metadata_file = False

    # Access the dataset folder
    for entry in os.scandir(datasets_dir):
        if entry.is_file():
            if entry.name == '__metadata__.json':
                has_seen_metadata_file = True

                # We have found the dataset's metadata file, build out information based on it
                with open(os.path.join(datasets_dir, entry.name)) as f:
                    metadata = json.load(f)
                id = metadata['id']
                name = metadata['name']
                item_count = metadata['itemCount']
                created_at = datetime.strptime(metadata['createdAt'], '%Y-%m-%dT%H:%M:%S.%fZ')
                accessed_at = datetime.strptime(metadata['accessedAt'], '%Y-%m-%dT%H:%M:%S.%fZ')
                modified_at = datetime.strptime(metadata['modifiedAt'], '%Y-%m-%dT%H:%M:%S.%fZ')

                continue

            with open(os.path.join(datasets_dir, entry.name)) as f:
                entry_content = json.load(f)
            entry_name = entry.name.split('.')[0]

            entries[entry_name] = entry_content

            if not has_seen_metadata_file:
                item_count += 1

    if id is None and name is None:
        is_uuid = _is_uuid(entry_name_or_id)

        if is_uuid:
            id = entry_name_or_id
        else:
            name = entry_name_or_id

    new_client = DatasetClient(base_storage_directory=client._datasets_directory, client=client, id=id, name=name)

    # Overwrite properties
    new_client._accessed_at = accessed_at
    new_client._created_at = created_at
    new_client._modified_at = modified_at
    new_client._item_count = item_count

    for entry_id, content in entries.items():
        new_client._dataset_entries[entry_id] = content

    client._datasets_handled.append(new_client)

    return new_client



import asyncio
import inspect
import os
import sys
import traceback
from datetime import datetime
from types import TracebackType
from typing import Any, Awaitable, Callable, Dict, List, Optional, Type, TypeVar, Union, cast

from apify_client import ApifyClientAsync
from apify_client.consts import WebhookEventType

from ._utils import (
    _fetch_and_parse_env_var,
    _get_cpu_usage_percent,
    _get_memory_usage_bytes,
    _is_running_in_ipython,
    _log_system_info,
    _run_func_at_interval_async,
    _wrap_internal,
    dualproperty,
)
from .config import Configuration
from .consts import EVENT_LISTENERS_TIMEOUT_SECS, ActorEventType, ActorExitCodes, ApifyEnvVars
from .event_manager import EventManager
from .memory_storage import MemoryStorage
from .proxy_configuration import ProxyConfiguration
from .storage_client_manager import StorageClientManager
from .storages import Dataset, KeyValueStore, RequestQueue, StorageManager

MainReturnType = TypeVar('MainReturnType')

# This metaclass is needed so you can do `async with Actor: ...` instead of `async with Actor() as a: ...`
# and have automatic `Actor.init()` and `Actor.exit()`
# TODO: decide if this mumbo jumbo is worth it or not, or if it maybe breaks something


class _ActorContextManager(type):
    @staticmethod
    async def __aenter__() -> Type['Actor']:
        await Actor.init()
        return Actor

    @staticmethod
    async def __aexit__(
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        if not Actor._get_default_instance()._is_exiting:
            if exc_type:
                await Actor.fail(
                    exit_code=ActorExitCodes.ERROR_USER_FUNCTION_THREW.value,
                    _exc_type=exc_type,
                    _exc_value=exc_value,
                    _exc_traceback=exc_traceback,
                )
            else:
                await Actor.exit()


class Actor(metaclass=_ActorContextManager):
    """Class representing an Apify Actor."""

    _default_instance: Optional['Actor'] = None
    _apify_client: ApifyClientAsync
    _memory_storage: MemoryStorage
    _config: Configuration
    _event_manager: EventManager
    _send_system_info_interval_task: Optional[asyncio.Task] = None
    _send_persist_state_interval_task: Optional[asyncio.Task] = None
    _is_exiting = False

    def __init__(self, config: Optional[Configuration] = None) -> None:
        """Create an Actor instance.

        Args:
            config (Configuration, optional): The actor configuration to be used. If not passed, a new Configuration instance will be created.
        """
        # To have methods which work the same as classmethods and instance methods,
        # so you can do both Actor.xxx() and Actor().xxx(),
        # we need to have an `_xxx_internal` instance method which contains the actual implementation of the method,
        # and then in the instance constructor overwrite the `xxx` classmethod with the `_xxx_internal` instance method,
        # while copying the annotations, types and so on.
        self.init = _wrap_internal(self._init_internal, self.init)  # type: ignore
        self.exit = _wrap_internal(self._exit_internal, self.exit)  # type: ignore
        self.fail = _wrap_internal(self._fail_internal, self.fail)  # type: ignore
        self.main = _wrap_internal(self._main_internal, self.main)  # type: ignore
        self.new_client = _wrap_internal(self._new_client_internal, self.new_client)  # type: ignore

        self.open_dataset = _wrap_internal(self._open_dataset_internal, self.open_dataset)  # type: ignore
        self.open_key_value_store = _wrap_internal(self._open_key_value_store_internal, self.open_key_value_store)  # type: ignore
        self.open_request_queue = _wrap_internal(self._open_request_queue_internal, self.open_request_queue)  # type: ignore
        self.push_data = _wrap_internal(self._push_data_internal, self.push_data)  # type: ignore
        self.get_input = _wrap_internal(self._get_input_internal, self.get_input)  # type: ignore
        self.get_value = _wrap_internal(self._get_value_internal, self.get_value)  # type: ignore
        self.set_value = _wrap_internal(self._set_value_internal, self.set_value)  # type: ignore

        self.on = _wrap_internal(self._on_internal, self.on)  # type: ignore
        self.off = _wrap_internal(self._off_internal, self.off)  # type: ignore

        self.is_at_home = _wrap_internal(self._is_at_home_internal, self.is_at_home)  # type: ignore
        self.get_env = _wrap_internal(self._get_env_internal, self.get_env)  # type: ignore

        self.start = _wrap_internal(self._start_internal, self.start)  # type: ignore
        self.call = _wrap_internal(self._call_internal, self.call)  # type: ignore
        self.call_task = _wrap_internal(self._call_task_internal, self.call_task)  # type: ignore
        self.abort = _wrap_internal(self._abort_internal, self.abort)  # type: ignore
        self.metamorph = _wrap_internal(self._metamorph_internal, self.metamorph)  # type: ignore
        self.reboot = _wrap_internal(self._reboot_internal, self.reboot)  # type: ignore
        self.add_webhook = _wrap_internal(self._add_webhook_internal, self.add_webhook)  # type: ignore
        self.set_status_message = _wrap_internal(self._set_status_message_internal, self.set_status_message)  # type: ignore
        self.create_proxy_configuration = _wrap_internal(self._create_proxy_configuration_internal, self.create_proxy_configuration)  # type: ignore

        self._config: Configuration = config or Configuration()
        self._apify_client = self.new_client()
        self._event_manager = EventManager(config=self._config)

        self._is_initialized = False

    async def __aenter__(self) -> 'Actor':
        """Initialize the Actor.

        Automatically initializes the Actor instance when you use it in an `async with ...` statement.

        When you exit the `async with` block, the `Actor.exit()` method is called,
        and if any exception happens while executing the block code,
        the `Actor.fail` method is called.
        """
        await self.init()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        """Exit the Actor, handling any exceptions properly.

        When you exit the `async with` block, the `Actor.exit()` method is called,
        and if any exception happens while executing the block code,
        the `Actor.fail` method is called.
        """
        if not self._is_exiting:
            if exc_type:
                await self.fail(
                    exit_code=ActorExitCodes.ERROR_USER_FUNCTION_THREW.value,
                    _exc_type=exc_type,
                    _exc_value=exc_value,
                    _exc_traceback=exc_traceback,
                )
            else:
                await self.exit()

    @classmethod
    def _get_default_instance(cls) -> 'Actor':
        if not cls._default_instance:
            cls._default_instance = cls(config=Configuration.get_global_configuration())

        return cls._default_instance

    @dualproperty
    def apify_client(self_or_cls) -> ApifyClientAsync:  # noqa: N805
        """The ApifyClientAsync instance the Actor instance uses."""  # noqa: D401
        if isinstance(self_or_cls, type):
            return self_or_cls._get_default_instance()._apify_client
        else:
            return self_or_cls._apify_client

    @dualproperty
    def config(self_or_cls) -> Configuration:  # noqa: N805
        """The Configuration instance the Actor instance uses."""  # noqa: D401
        if isinstance(self_or_cls, type):
            return self_or_cls._get_default_instance()._config
        else:
            return self_or_cls._config

    @dualproperty
    def event_manager(self_or_cls) -> EventManager:  # noqa: N805
        """The EventManager instance the Actor instance uses."""  # noqa: D401
        if isinstance(self_or_cls, type):
            return self_or_cls._get_default_instance()._event_manager
        else:
            return self_or_cls._event_manager

    def _raise_if_not_initialized(self) -> None:
        if not self._is_initialized:
            raise RuntimeError('The actor was not initialized!')

    @classmethod
    async def init(cls) -> None:
        """Initialize the actor instance.

        This initializes the Actor instance.
        It configures the right storage client based on whether the actor is running locally or on the Apify platform,
        it initializes the event manager for processing actor events,
        and starts an interval for regularly sending `PERSIST_STATE` events,
        so that the actor can regularly persist its state in response to these events.

        This method should be called immediately before performing any additional actor actions,
        and it should be called only once.
        """
        return await cls._get_default_instance().init()

    async def _init_internal(self) -> None:
        if self._is_initialized:
            raise RuntimeError('The actor was already initialized!')

        self._is_exiting = False

        print('Initializing actor...')
        _log_system_info()

        # TODO: Print outdated SDK version warning (we need a new env var for this)

        await self._event_manager.init()

        self._send_persist_state_interval_task = asyncio.create_task(
            _run_func_at_interval_async(
                lambda: self._event_manager.emit(ActorEventType.PERSIST_STATE, {'isMigrating': False}),
                self._config.persist_state_interval_millis / 1000,
            ),
        )

        if self.is_at_home():
            StorageClientManager.set_storage_client(self._apify_client)
        else:
            self._send_system_info_interval_task = asyncio.create_task(
                _run_func_at_interval_async(
                    lambda: self._event_manager.emit(ActorEventType.SYSTEM_INFO, self._get_system_info()),
                    self._config.system_info_interval_millis / 1000,
                ),
            )

        self._event_manager.on(ActorEventType.MIGRATING, self._respond_to_migrating_event)

        # The CPU usage is calculated as an average between two last calls to psutil
        # We need to make a first, dummy call, so the next calls have something to compare itself agains
        _get_cpu_usage_percent()

        self._is_initialized = True

    def _get_system_info(self) -> Dict:
        cpu_usage_percent = _get_cpu_usage_percent()
        memory_usage_bytes = _get_memory_usage_bytes()
        # This is in camel case to be compatible with the events from the platform
        result = {
            'createdAt': datetime.utcnow().isoformat(timespec='milliseconds') + 'Z',
            'cpuCurrentUsage': cpu_usage_percent,
            'memCurrentBytes': memory_usage_bytes,
        }
        if self._config.max_used_cpu_ratio:
            result['isCpuOverloaded'] = (cpu_usage_percent > 100 * self._config.max_used_cpu_ratio)

        return result

    def _respond_to_migrating_event(self) -> None:
        # Don't emit any more regular persist state events
        if self._send_persist_state_interval_task:
            self._send_persist_state_interval_task.cancel()
        self._event_manager.emit(ActorEventType.PERSIST_STATE, {'is_migrating': True})

    async def _cancel_event_emitting_intervals(self) -> None:
        if self._send_persist_state_interval_task and not self._send_persist_state_interval_task.cancelled():
            self._send_persist_state_interval_task.cancel()
            try:
                await self._send_persist_state_interval_task
            except asyncio.CancelledError:
                pass

        if self._send_system_info_interval_task and not self._send_system_info_interval_task.cancelled():
            self._send_system_info_interval_task.cancel()
            try:
                await self._send_system_info_interval_task
            except asyncio.CancelledError:
                pass

    @classmethod
    async def exit(
        cls,
        *,
        exit_code: int = 0,
        event_listeners_timeout_secs: Optional[int] = EVENT_LISTENERS_TIMEOUT_SECS,
    ) -> None:
        """Exit the actor instance.

        This stops the Actor instance.
        It cancels all the intervals for regularly sending `PERSIST_STATE` events,
        sends a final `PERSIST_STATE` event,
        waits for all the event listeners to finish,
        and stops the event manager.

        Args:
            exit_code (int, optional): The exit code with which the actor should fail (defaults to `0`).
            event_listeners_timeout_secs (int, optional): How long should the actor wait for actor event listeners to finish before exiting
        """
        return await cls._get_default_instance().exit(
            exit_code=exit_code,
            event_listeners_timeout_secs=event_listeners_timeout_secs,
        )

    async def _exit_internal(
        self,
        *,
        exit_code: int = 0,
        event_listeners_timeout_secs: Optional[int] = EVENT_LISTENERS_TIMEOUT_SECS,
    ) -> None:
        self._raise_if_not_initialized()

        self._is_exiting = True

        print(f'Exiting actor with exit code {exit_code}')

        await self._cancel_event_emitting_intervals()

        # Send final persist state event
        self._event_manager.emit(ActorEventType.PERSIST_STATE, {'isMigrating': False})

        # Sleep for a bit so that the listeners have a chance to trigger,
        await asyncio.sleep(0.1)

        await self._event_manager.close(event_listeners_timeout_secs=event_listeners_timeout_secs)

        # TODO: once we have in-memory storage, teardown the in-memory storage client here

        self._is_initialized = False

        if not _is_running_in_ipython() and not os.getenv('PYTEST_CURRENT_TEST', False):
            sys.exit(exit_code)
        else:
            print(f'Not calling sys.exit({exit_code}) because actor is running in IPython')

    @classmethod
    async def fail(
        cls,
        *,
        exit_code: int = 1,
        _exc_type: Optional[Type[BaseException]] = None,
        _exc_value: Optional[BaseException] = None,
        _exc_traceback: Optional[TracebackType] = None,
    ) -> None:
        """Fail the actor instance.

        This performs all the same steps as Actor.exit(),
        but it additionally sets the exit code to `1` (by default).

        Args:
            exit_code (int, optional): The exit code with which the actor should fail (defaults to `1`).
        """
        return await cls._get_default_instance().fail(
            exit_code=exit_code,
            _exc_type=_exc_type,
            _exc_value=_exc_value,
            _exc_traceback=_exc_traceback,
        )

    async def _fail_internal(
        self,
        *,
        exit_code: int = 1,
        _exc_type: Optional[Type[BaseException]] = None,
        _exc_value: Optional[BaseException] = None,
        _exc_traceback: Optional[TracebackType] = None,
    ) -> None:
        self._raise_if_not_initialized()

        # In IPython, we don't run `sys.exit()` during actor exits,
        # so the exception traceback will be printed on its own
        if _exc_type and not _is_running_in_ipython():
            print('Actor failed with an exception:')
            traceback.print_exception(_exc_type, _exc_value, _exc_traceback)

        await self.exit(exit_code=exit_code)

    @classmethod
    async def main(cls, main_actor_function: Callable[[], MainReturnType]) -> Optional[MainReturnType]:
        """Initialize the actor, run the passed function and finish the actor cleanly.

        **The `Actor.main()` function is optional** and is provided merely for your convenience.
        It is mainly useful when you're running your code as an actor on the [Apify platform](https://apify.com/actors).

        The `Actor.main()` function performs the following actions:

        - When running on the Apify platform (i.e. `APIFY_IS_AT_HOME` environment variable is set),
          it sets up a connection to listen for platform events.
          For example, to get a notification about an imminent migration to another server.
        - It invokes the user function passed as the `main_actor_function` parameter.
        - If the user function was an async function, it awaits it.
        - If the user function throws an exception or some other error is encountered,
          it prints error details to console so that they are stored to the log,
          and finishes the actor cleanly.
        - Finally, it exits the Python process, with zero exit code on success and non-zero on errors.

        Args:
            main_actor_function (Callable): The user function which should be run in the actor
        """
        return await cls._get_default_instance().main(
            main_actor_function=main_actor_function,
        )

    async def _main_internal(self, main_actor_function: Callable[[], MainReturnType]) -> Optional[MainReturnType]:
        if not inspect.isfunction(main_actor_function):
            raise TypeError(f'First argument passed to Actor.main() must be a function, but instead it was {type(main_actor_function)}')

        await self.init()
        try:
            if inspect.iscoroutinefunction(main_actor_function):
                res = await main_actor_function()
            else:
                res = main_actor_function()
            await self.exit()
            return cast(MainReturnType, res)
        except Exception as e:
            await self.fail(
                exit_code=ActorExitCodes.ERROR_USER_FUNCTION_THREW.value,
                _exc_type=type(e),
                _exc_value=e,
                _exc_traceback=e.__traceback__,
            )
        return None

    @classmethod
    def new_client(
        cls,
        *,
        token: Optional[str] = None,
        api_url: Optional[str] = None,
        max_retries: Optional[int] = None,
        min_delay_between_retries_millis: Optional[int] = None,
        timeout_secs: Optional[int] = None,
    ) -> ApifyClientAsync:
        """Return a new instance of the Apify API client.

        The `ApifyClientAsync` class is provided by the [apify-client](https://github.com/apify/apify-client-python) package,
        and it is automatically configured using the `APIFY_API_BASE_URL` and `APIFY_TOKEN` environment variables.

        You can override the token via the available options.
        That's useful if you want to use the client as a different Apify user than the SDK internals are using.

        Args:
            token (str, optional): The Apify API token
            api_url (str, optional): The URL of the Apify API server to which to connect to. Defaults to https://api.apify.com
            max_retries (int, optional): How many times to retry a failed request at most
            min_delay_between_retries_millis (int, optional): How long will the client wait between retrying requests
                (increases exponentially from this value)
            timeout_secs (int, optional): The socket timeout of the HTTP requests sent to the Apify API
        """
        return cls._get_default_instance().new_client(
            token=token,
            api_url=api_url,
            max_retries=max_retries,
            min_delay_between_retries_millis=min_delay_between_retries_millis,
            timeout_secs=timeout_secs,
        )

    def _new_client_internal(
        self,
        *,
        token: Optional[str] = None,
        api_url: Optional[str] = None,
        max_retries: Optional[int] = None,
        min_delay_between_retries_millis: Optional[int] = None,
        timeout_secs: Optional[int] = None,
    ) -> ApifyClientAsync:
        token = token or self._config.token
        api_url = api_url or self._config.api_base_url
        return ApifyClientAsync(
            token=token,
            api_url=api_url,
            max_retries=max_retries,
            min_delay_between_retries_millis=min_delay_between_retries_millis,
            timeout_secs=timeout_secs,
        )

    def _get_storage_client(self, force_cloud: bool) -> Optional[ApifyClientAsync]:
        return self._apify_client if force_cloud else None

    @classmethod
    async def open_dataset(cls, dataset_id_or_name: Optional[str] = None, *, force_cloud: bool = False) -> Dataset:
        """TODO: docs."""
        return await cls._get_default_instance().open_dataset(dataset_id_or_name=dataset_id_or_name, force_cloud=force_cloud)

    async def _open_dataset_internal(self, dataset_id_or_name: Optional[str] = None, *, force_cloud: bool = False) -> Dataset:
        self._raise_if_not_initialized()

        return await StorageManager.open_storage(Dataset, dataset_id_or_name, self._get_storage_client(force_cloud), self._config)

    @classmethod
    async def open_key_value_store(cls, key_value_store_id_or_name: Optional[str] = None, *, force_cloud: bool = False) -> KeyValueStore:
        """TODO: docs."""
        return await cls._get_default_instance().open_key_value_store(key_value_store_id_or_name=key_value_store_id_or_name, force_cloud=force_cloud)

    async def _open_key_value_store_internal(self, key_value_store_id_or_name: Optional[str] = None, *, force_cloud: bool = False) -> KeyValueStore:
        self._raise_if_not_initialized()

        return await StorageManager.open_storage(KeyValueStore, key_value_store_id_or_name, self._get_storage_client(force_cloud), self._config)

    @classmethod
    async def open_request_queue(cls, request_queue_id_or_name: Optional[str] = None, *, force_cloud: bool = False) -> RequestQueue:
        """TODO: docs."""
        return await cls._get_default_instance().open_request_queue(request_queue_id_or_name=request_queue_id_or_name, force_cloud=force_cloud)

    async def _open_request_queue_internal(
        self,
        request_queue_id_or_name: Optional[str] = None,
        *,
        force_cloud: bool = False,
    ) -> RequestQueue:
        self._raise_if_not_initialized()

        return await StorageManager.open_storage(RequestQueue, request_queue_id_or_name, self._get_storage_client(force_cloud), self._config)

    @classmethod
    async def push_data(cls, data: Any) -> None:
        """Store an object or a list of objects to the default dataset of the current actor run.

        Args:
            data (object or list of objects, optional): The data to push to the default dataset.
        """
        return await cls._get_default_instance().push_data(data=data)

    async def _push_data_internal(self, data: Any) -> None:
        self._raise_if_not_initialized()

        if not data:
            return

        if not isinstance(data, list):  # TODO: Memory storage does this on its own...
            data = [data]

        dataset = await self.open_dataset()
        await dataset.push_data(data)

    @classmethod
    async def get_input(cls) -> Any:
        """Get the actor input value from the default key-value store associated with the current actor run."""
        return await cls._get_default_instance().get_input()

    async def _get_input_internal(self) -> Any:
        self._raise_if_not_initialized()

        # TODO: decryption

        return await self.get_value(self._config.input_key)

    @classmethod
    async def get_value(cls, key: str) -> Any:
        """Get a value from the default key-value store associated with the current actor run.

        Args:
            key (str): The key of the record which to retrieve.
        """
        return await cls._get_default_instance().get_value(key=key)

    async def _get_value_internal(self, key: str) -> Any:
        self._raise_if_not_initialized()

        key_value_store = await self.open_key_value_store()
        value = await key_value_store.get_value(key)
        return value

    @classmethod
    async def set_value(
        cls,
        key: str,
        value: Any,
        *,
        content_type: Optional[str] = None,
    ) -> None:
        """Set or delete a value in the default key-value store associated with the current actor run.

        Args:
            key (str): The key of the record which to set.
            value (any): The value of the record which to set, or None, if the record should be deleted.
            content_type (str, optional): The content type which should be set to the value.
        """
        return await cls._get_default_instance().set_value(
            key=key,
            value=value,
            content_type=content_type,
        )

    async def _set_value_internal(
        self,
        key: str,
        value: Any,
        *,
        content_type: Optional[str] = None,
    ) -> None:
        self._raise_if_not_initialized()

        key_value_store = await self.open_key_value_store()
        return await key_value_store.set_value(key, value, content_type=content_type)

    @classmethod
    def on(cls, event_name: ActorEventType, listener: Callable) -> Callable:
        """Add an event listener to the actor's event manager.

        The following events can be emitted:
         - `ActorEventType.SYSTEM_INFO`:
            Emitted every minute, the event data contains info about the resource usage of the actor.
         - `ActorEventType.MIGRATING`:
            Emitted when the actor running on the Apify platform is going to be migrated to another worker server soon.
            You can use it to persist the state of the actor and abort the run, to speed up the migration.
         - `ActorEventType.PERSIST_STATE`:
            Emitted in regular intervals (by default 60 seconds) to notify the actor that it should persist its state,
            in order to avoid repeating all work when the actor restarts.
            This event is automatically emitted together with the migrating event,
            in which case the `isMigrating` flag in the event data is set to True, otherwise the flag is False.
            Note that this event is provided merely for your convenience,
            you can achieve the same effect using an interval and listening for the migrating event.
         - `ActorEventType.ABORTING`:
            When a user aborts an actor run on the Apify platform,
            they can choose to abort it gracefully, to allow the actor some time before getting terminated.
            This graceful abort emits the aborting event, which you can use to clean up the actor state.

        Args:
            event_name (ActorEventType): The actor event for which to listen to.
            listener (Callable): The function which is to be called when the event is emitted (can be async).
        """
        return cls._get_default_instance().on(event_name, listener)

    def _on_internal(self, event_name: ActorEventType, listener: Callable) -> Callable:
        self._raise_if_not_initialized()

        return self._event_manager.on(event_name, listener)

    @classmethod
    def off(cls, event_name: ActorEventType, listener: Optional[Callable] = None) -> None:
        """Remove a listener, or all listeners, from an actor event.

        Args:
            event_name (ActorEventType): The actor event for which to remove listeners.
            listener (Callable, optional): The listener which is supposed to be removed. If not passed, all listeners of this event are removed.
        """
        return cls._get_default_instance().off(event_name, listener)

    def _off_internal(self, event_name: ActorEventType, listener: Optional[Callable] = None) -> None:
        self._raise_if_not_initialized()

        return self._event_manager.off(event_name, listener)

    @classmethod
    def is_at_home(cls) -> bool:
        """Return `True` when the actor is running on the Apify platform, and `False` otherwise (for example when running locally)."""
        return cls._get_default_instance().is_at_home()

    def _is_at_home_internal(self) -> bool:
        return self._config.is_at_home

    @classmethod
    def get_env(cls) -> Dict:
        """Return a dictionary with information parsed from all the `APIFY_XXX` environment variables.

        For a list of all the environment variables,
        see the [Actor documentation](https://docs.apify.com/actor/run#environment-variables).
        If some variables are not defined or are invalid, the corresponding value in the resulting dictionary will be None.
        """
        return cls._get_default_instance().get_env()

    def _get_env_internal(self) -> Dict:
        self._raise_if_not_initialized()

        return {
            env_var.name.lower(): _fetch_and_parse_env_var(env_var) for env_var in ApifyEnvVars
        }

    @classmethod
    async def start(
        cls,
        actor_id: str,
        run_input: Optional[Any] = None,
        *,
        token: Optional[str] = None,
        content_type: Optional[str] = None,
        build: Optional[str] = None,
        memory_mbytes: Optional[int] = None,
        timeout_secs: Optional[int] = None,
        wait_for_finish: Optional[int] = None,
        webhooks: Optional[List[Dict]] = None,
    ) -> Dict:
        """Run an actor on the Apify platform.

        Unlike `Actor.call`, this method just starts the run without waiting for finish.

        Args:
            actor_id (str): The ID of the actor to be run.
            run_input (Any, optional): The input to pass to the actor run.
            token (str, optional): The Apify API token to use for this request (defaults to the `APIFY_TOKEN` environment variable).
            content_type (str, optional): The content type of the input.
            build (str, optional): Specifies the actor build to run. It can be either a build tag or build number.
                                   By default, the run uses the build specified in the default run configuration for the actor (typically latest).
            memory_mbytes (int, optional): Memory limit for the run, in megabytes.
                                           By default, the run uses a memory limit specified in the default run configuration for the actor.
            timeout_secs (int, optional): Optional timeout for the run, in seconds.
                                          By default, the run uses timeout specified in the default run configuration for the actor.
            wait_for_finish (int, optional): The maximum number of seconds the server waits for the run to finish.
                                               By default, it is 0, the maximum value is 300.
            webhooks (list of dict, optional): Optional ad-hoc webhooks (https://docs.apify.com/webhooks/ad-hoc-webhooks)
                                               associated with the actor run which can be used to receive a notification,
                                               e.g. when the actor finished or failed.
                                               If you already have a webhook set up for the actor or task, you do not have to add it again here.
                                               Each webhook is represented by a dictionary containing these items:
                                               * ``event_types``: list of ``WebhookEventType`` values which trigger the webhook
                                               * ``request_url``: URL to which to send the webhook HTTP request
                                               * ``payload_template`` (optional): Optional template for the request payload

        Returns:
            dict: Info about the started actor run
        """
        return await cls._get_default_instance().start(
            actor_id=actor_id,
            run_input=run_input,
            token=token,
            content_type=content_type,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=timeout_secs,
            wait_for_finish=wait_for_finish,
            webhooks=webhooks,
        )

    async def _start_internal(
        self,
        actor_id: str,
        run_input: Optional[Any] = None,
        *,
        token: Optional[str] = None,
        content_type: Optional[str] = None,
        build: Optional[str] = None,
        memory_mbytes: Optional[int] = None,
        timeout_secs: Optional[int] = None,
        wait_for_finish: Optional[int] = None,
        webhooks: Optional[List[Dict]] = None,
    ) -> Dict:
        self._raise_if_not_initialized()

        if token:
            client = self.new_client(token=token)
        else:
            client = self._apify_client

        return await client.actor(actor_id).start(
            run_input=run_input,
            content_type=content_type,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=timeout_secs,
            wait_for_finish=wait_for_finish,
            webhooks=webhooks,
        )

    @classmethod
    async def abort(
        cls,
        run_id: str,
        *,
        token: Optional[str] = None,
        gracefully: Optional[bool] = None,
    ) -> Dict:
        """Abort given actor run on the Apify platform using the current user account (determined by the `APIFY_TOKEN` environment variable).

        Args:
            run_id (str): The ID of the actor run to be aborted.
            token (str, optional): The Apify API token to use for this request (defaults to the `APIFY_TOKEN` environment variable).
            gracefully (bool, optional): If True, the actor run will abort gracefully.
                It will send ``aborting`` and ``persistStates`` events into the run and force-stop the run after 30 seconds.
                It is helpful in cases where you plan to resurrect the run later.

        Returns:
            dict: Info about the aborted actor run
        """
        return await cls._get_default_instance().abort(
            run_id=run_id,
            token=token,
            gracefully=gracefully,
        )

    async def _abort_internal(
        self,
        run_id: str,
        *,
        token: Optional[str] = None,
        status_message: Optional[str] = None,
        gracefully: Optional[bool] = None,
    ) -> Dict:
        self._raise_if_not_initialized()

        if token:
            client = self.new_client(token=token)
        else:
            client = self._apify_client

        if status_message:
            await client.run(run_id).update(status_message=status_message)

        return await client.run(run_id).abort(gracefully=gracefully)

    @classmethod
    async def call(
        cls,
        actor_id: str,
        run_input: Optional[Any] = None,
        *,
        token: Optional[str] = None,
        content_type: Optional[str] = None,
        build: Optional[str] = None,
        memory_mbytes: Optional[int] = None,
        timeout_secs: Optional[int] = None,
        webhooks: Optional[List[Dict]] = None,
        wait_secs: Optional[int] = None,
    ) -> Optional[Dict]:
        """Start an actor on the Apify Platform and wait for it to finish before returning.

        It waits indefinitely, unless the wait_secs argument is provided.

        Args:
            actor_id (str): The ID of the actor to be run.
            run_input (Any, optional): The input to pass to the actor run.
            token (str, optional): The Apify API token to use for this request (defaults to the `APIFY_TOKEN` environment variable).
            content_type (str, optional): The content type of the input.
            build (str, optional): Specifies the actor build to run. It can be either a build tag or build number.
                                   By default, the run uses the build specified in the default run configuration for the actor (typically latest).
            memory_mbytes (int, optional): Memory limit for the run, in megabytes.
                                           By default, the run uses a memory limit specified in the default run configuration for the actor.
            timeout_secs (int, optional): Optional timeout for the run, in seconds.
                                          By default, the run uses timeout specified in the default run configuration for the actor.
            webhooks (list, optional): Optional webhooks (https://docs.apify.com/webhooks) associated with the actor run,
                                       which can be used to receive a notification, e.g. when the actor finished or failed.
                                       If you already have a webhook set up for the actor, you do not have to add it again here.
            wait_secs (int, optional): The maximum number of seconds the server waits for the run to finish. If not provided, waits indefinitely.

        Returns:
            dict: Info about the started actor run
        """
        return await cls._get_default_instance().call(
            actor_id=actor_id,
            token=token,
            run_input=run_input,
            content_type=content_type,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=timeout_secs,
            webhooks=webhooks,
            wait_secs=wait_secs,
        )

    async def _call_internal(
        self,
        actor_id: str,
        run_input: Optional[Any] = None,
        *,
        token: Optional[str] = None,
        content_type: Optional[str] = None,
        build: Optional[str] = None,
        memory_mbytes: Optional[int] = None,
        timeout_secs: Optional[int] = None,
        webhooks: Optional[List[Dict]] = None,
        wait_secs: Optional[int] = None,
    ) -> Optional[Dict]:
        self._raise_if_not_initialized()

        if token:
            client = self.new_client(token=token)
        else:
            client = self._apify_client

        return await client.actor(actor_id).call(
            run_input=run_input,
            content_type=content_type,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=timeout_secs,
            webhooks=webhooks,
            wait_secs=wait_secs,
        )

    @classmethod
    async def call_task(
        cls,
        task_id: str,
        task_input: Optional[Dict[str, Any]] = None,
        *,
        build: Optional[str] = None,
        memory_mbytes: Optional[int] = None,
        timeout_secs: Optional[int] = None,
        webhooks: Optional[List[Dict]] = None,
        wait_secs: Optional[int] = None,
        token: Optional[str] = None,
    ) -> Optional[Dict]:
        """Start an actor task on the Apify Platform and wait for it to finish before returning.

        It waits indefinitely, unless the wait_secs argument is provided.

        Note that an actor task is a saved input configuration and options for an actor.
        If you want to run an actor directly rather than an actor task, please use the `Actor.call`

        Args:
            task_id (str): The ID of the actor to be run.
            task_input (Any, optional): Overrides the input to pass to the actor run.
            token (str, optional): The Apify API token to use for this request (defaults to the `APIFY_TOKEN` environment variable).
            content_type (str, optional): The content type of the input.
            build (str, optional): Specifies the actor build to run. It can be either a build tag or build number.
                                   By default, the run uses the build specified in the default run configuration for the actor (typically latest).
            memory_mbytes (int, optional): Memory limit for the run, in megabytes.
                                           By default, the run uses a memory limit specified in the default run configuration for the actor.
            timeout_secs (int, optional): Optional timeout for the run, in seconds.
                                          By default, the run uses timeout specified in the default run configuration for the actor.
            webhooks (list, optional): Optional webhooks (https://docs.apify.com/webhooks) associated with the actor run,
                                       which can be used to receive a notification, e.g. when the actor finished or failed.
                                       If you already have a webhook set up for the actor, you do not have to add it again here.
            wait_secs (int, optional): The maximum number of seconds the server waits for the run to finish. If not provided, waits indefinitely.

        Returns:
            dict: Info about the started actor run
        """
        return await cls._get_default_instance().call_task(
            task_id=task_id,
            task_input=task_input,
            token=token,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=timeout_secs,
            webhooks=webhooks,
            wait_secs=wait_secs,
        )

    async def _call_task_internal(
        self,
        task_id: str,
        task_input: Optional[Dict[str, Any]] = None,
        *,
        build: Optional[str] = None,
        memory_mbytes: Optional[int] = None,
        timeout_secs: Optional[int] = None,
        webhooks: Optional[List[Dict]] = None,
        wait_secs: Optional[int] = None,
        token: Optional[str] = None,
    ) -> Optional[Dict]:
        self._raise_if_not_initialized()

        if token:
            client = self.new_client(token=token)
        else:
            client = self._apify_client

        return await client.task(task_id).call(
            task_input=task_input,
            build=build,
            memory_mbytes=memory_mbytes,
            timeout_secs=timeout_secs,
            webhooks=webhooks,
            wait_secs=wait_secs,
        )

    @classmethod
    async def metamorph(
        cls,
        target_actor_id: str,
        run_input: Optional[Any] = None,
        *,
        target_actor_build: Optional[str] = None,
        content_type: Optional[str] = None,
        custom_after_sleep_millis: Optional[int] = None,
    ) -> None:
        """Transform this actor run to an actor run of a different actor.

        The platform stops the current actor container and starts a new container with the new actor instead.
        All the default storages are preserved,
        and the new input is stored under the `INPUT-METAMORPH-1` key in the same default key-value store.

        Args:
            target_actor_id (str): ID of the target actor that the run should be transformed into
            run_input (Any, optional): The input to pass to the new run.
            target_actor_build (str, optional): The build of the target actor. It can be either a build tag or build number.
                By default, the run uses the build specified in the default run configuration for the target actor (typically the latest build).
            content_type (str, optional): The content type of the input.
            custom_after_sleep_millis (int, optional): How long to sleep for after the metamorph, to wait for the container to be stopped.

        Returns:
            dict: The actor run data.
        """
        return await cls._get_default_instance().metamorph(
            target_actor_id=target_actor_id,
            target_actor_build=target_actor_build,
            run_input=run_input,
            content_type=content_type,
            custom_after_sleep_millis=custom_after_sleep_millis,
        )

    async def _metamorph_internal(
        self,
        target_actor_id: str,
        run_input: Optional[Any] = None,
        *,
        target_actor_build: Optional[str] = None,
        content_type: Optional[str] = None,
        custom_after_sleep_millis: Optional[int] = None,
    ) -> None:
        self._raise_if_not_initialized()

        if not self.is_at_home():
            print('Actor.metamorph() is only supported when running on the Apify platform.')
            return

        if not custom_after_sleep_millis:
            custom_after_sleep_millis = self._config.metamorph_after_sleep_millis

        # If is_at_home() is True, config.actor_run_id is always set
        assert self._config.actor_run_id is not None

        await self._apify_client.run(self._config.actor_run_id).metamorph(
            target_actor_id=target_actor_id,
            run_input=run_input,
            target_actor_build=target_actor_build,
            content_type=content_type,
        )

        if custom_after_sleep_millis:
            await asyncio.sleep(custom_after_sleep_millis / 1000)

    @classmethod
    async def reboot(
        cls,
        *,
        event_listeners_timeout_secs: Optional[int] = EVENT_LISTENERS_TIMEOUT_SECS,
    ) -> None:
        """Internally reboot this actor.

        The system stops the current container and starts a new one, with the same run ID and default storages.

        Args:
            event_listeners_timeout_secs (int, optional): How long should the actor wait for actor event listeners to finish before exiting
        """
        return await cls._get_default_instance().reboot(event_listeners_timeout_secs=event_listeners_timeout_secs)

    async def _reboot_internal(
        self,
        *,
        event_listeners_timeout_secs: Optional[int] = EVENT_LISTENERS_TIMEOUT_SECS,
    ) -> None:
        self._raise_if_not_initialized()

        if not self.is_at_home():
            print('Actor.reboot() is only supported when running on the Apify platform.')
            return

        await self._cancel_event_emitting_intervals()

        self._event_manager.emit(ActorEventType.PERSIST_STATE, {'isMigrating': True})

        await self._event_manager.close(event_listeners_timeout_secs=event_listeners_timeout_secs)

        # If is_at_home() is True, config.actor_id is always set
        assert self._config.actor_id is not None

        await self.metamorph(self._config.actor_id)

    @classmethod
    async def add_webhook(
        cls,
        *,
        event_types: List[WebhookEventType],
        request_url: str,
        payload_template: Optional[str] = None,
        ignore_ssl_errors: Optional[bool] = None,
        do_not_retry: Optional[bool] = None,
        idempotency_key: Optional[str] = None,
    ) -> Dict:
        """Create an ad-hoc webhook for the current actor run.

        This webhook lets you receive a notification when the actor run finished or failed.

        Note that webhooks are only supported for actors running on the Apify platform.
        When running the actor locally, the function will print a warning and have no effect.

        For more information about Apify actor webhooks, please see the [documentation](https://docs.apify.com/webhooks).

        Args:
            event_types (list of WebhookEventType): List of event types that should trigger the webhook. At least one is required.
            request_url (str): URL that will be invoked once the webhook is triggered.
            payload_template (str, optional): Specification of the payload that will be sent to request_url
            ignore_ssl_errors (bool, optional): Whether the webhook should ignore SSL errors returned by request_url
            do_not_retry (bool, optional): Whether the webhook should retry sending the payload to request_url upon
                                           failure.
            idempotency_key (str, optional): A unique identifier of a webhook. You can use it to ensure that you won't
                                             create the same webhook multiple times.

        Returns:
            dict: The created webhook
        """
        return await cls._get_default_instance().add_webhook(
            event_types=event_types,
            request_url=request_url,
            payload_template=payload_template,
            ignore_ssl_errors=ignore_ssl_errors,
            do_not_retry=do_not_retry,
            idempotency_key=idempotency_key,
        )

    async def _add_webhook_internal(
        self,
        *,
        event_types: List[WebhookEventType],
        request_url: str,
        payload_template: Optional[str] = None,
        ignore_ssl_errors: Optional[bool] = None,
        do_not_retry: Optional[bool] = None,
        idempotency_key: Optional[str] = None,
    ) -> Optional[Dict]:
        self._raise_if_not_initialized()

        if not self.is_at_home():
            print('Actor.add_webhook() is only supported when running on the Apify platform.')
            return None

        # If is_at_home() is True, config.actor_run_id is always set
        assert self._config.actor_run_id is not None

        return await self._apify_client.webhooks().create(
            actor_run_id=self._config.actor_run_id,
            event_types=event_types,
            request_url=request_url,
            payload_template=payload_template,
            ignore_ssl_errors=ignore_ssl_errors,
            do_not_retry=do_not_retry,
            idempotency_key=idempotency_key,
        )

    @classmethod
    async def set_status_message(cls, status_message: str) -> Optional[Dict]:
        """Set the status message for the current actor run.

        Args:
            status_message (str): The status message to set to the run.

        Returns:
            dict: The updated actor run object
        """
        return await cls._get_default_instance().set_status_message(status_message=status_message)

    async def _set_status_message_internal(self, status_message: str) -> Optional[Dict]:
        self._raise_if_not_initialized()

        if not self.is_at_home():
            print('Actor.set_status_message() is only supported when running on the Apify platform.')
            return None

        # If is_at_home() is True, config.actor_run_id is always set
        assert self._config.actor_run_id is not None

        return await self._apify_client.run(self._config.actor_run_id).update(status_message=status_message)

    @classmethod
    async def create_proxy_configuration(
        cls,
        *,
        password: Optional[str] = None,
        groups: Optional[List[str]] = None,
        country_code: Optional[str] = None,
        proxy_urls: Optional[List[str]] = None,
        new_url_function: Optional[Union[Callable[[Optional[str]], str], Callable[[Optional[str]], Awaitable[str]]]] = None,
        actor_proxy_input: Optional[Dict] = None,  # this is the raw proxy input from the actor run input, it is not spread or snake_cased in here
    ) -> Optional[ProxyConfiguration]:
        """Create a ProxyConfiguration object with the passed proxy configuration.

        Configures connection to a proxy server with the provided options.
        Proxy servers are used to prevent target websites from blocking your crawlers based on IP address rate limits or blacklists.

        For more details and code examples, see the `ProxyConfiguration` class.

        Args:
            password (str, optional): Password for the Apify Proxy. If not provided, will use os.environ['APIFY_PROXY_PASSWORD'], if available.
            groups (list of str, optional): Proxy groups which the Apify Proxy should use, if provided.
            country_code (str, optional): Country which the Apify Proxy should use, if provided.
            proxy_urls (list of str, optional): Custom proxy server URLs which should be rotated through.
            new_url_function (Callable, optional): Function which returns a custom proxy URL to be used.
            actor_proxy_input (dict, optional): Proxy configuration field from the actor input, if actor has such input field.

        Returns:
            ProxyConfiguration, optional: ProxyConfiguration object with the passed configuration,
                                          or None, if no proxy should be used based on the configuration.
        """
        return await cls._get_default_instance().create_proxy_configuration(
            password=password,
            groups=groups,
            country_code=country_code,
            proxy_urls=proxy_urls,
            new_url_function=new_url_function,
            actor_proxy_input=actor_proxy_input,
        )

    async def _create_proxy_configuration_internal(
        self,
        *,
        password: Optional[str] = None,
        groups: Optional[List[str]] = None,
        country_code: Optional[str] = None,
        proxy_urls: Optional[List[str]] = None,
        new_url_function: Optional[Union[Callable[[Optional[str]], str], Callable[[Optional[str]], Awaitable[str]]]] = None,
        actor_proxy_input: Optional[Dict] = None,  # this is the raw proxy input from the actor run input, it is not spread or snake_cased in here
    ) -> Optional[ProxyConfiguration]:
        self._raise_if_not_initialized()

        if actor_proxy_input is not None:
            if actor_proxy_input.get('useApifyProxy', False):
                country_code = country_code or actor_proxy_input.get('apifyProxyCountry', None)
                groups = groups or actor_proxy_input.get('apifyProxyGroups', None)
            else:
                proxy_urls = actor_proxy_input.get('proxyUrls', [])
                if not proxy_urls:
                    return None

        proxy_configuration = ProxyConfiguration(
            password=password,
            groups=groups,
            country_code=country_code,
            proxy_urls=proxy_urls,
            new_url_function=new_url_function,
            actor_config=self._config,
            apify_client=self._apify_client,
        )

        await proxy_configuration.initialize()

        return proxy_configuration



import inspect
import re
from typing import Any, Awaitable, Callable, Dict, List, Optional, TypedDict, Union
from urllib.parse import urlparse

import httpx
from pydantic import AnyUrl, Field, validate_arguments
from typing_extensions import Annotated, NotRequired

from apify_client import ApifyClientAsync

from .config import Configuration
from .consts import ApifyEnvVars

APIFY_PROXY_VALUE_REGEX = r'^[\w._~]+$'
COUNTRY_CODE_REGEX = r'^[A-Z]{2}$'
SESSION_ID_MAX_LENGTH = 50
SESSION_ID_PYDANTIC_FIELD = Field(regex=APIFY_PROXY_VALUE_REGEX, max_length=SESSION_ID_MAX_LENGTH)


# TODO: Remove this once Pydantic fixes https://github.com/pydantic/pydantic/issues/4333
def _check(
    value: Any,
    *,
    label: Optional[str],
    regex: Optional[str] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
) -> None:
    error_str = f'Value {value}'
    if label:
        error_str += f' of argument {label}'

    if min_length and len(value) < min_length:
        raise ValueError(f'{error_str} is shorter than minimum allowed length {min_length}')

    if max_length and len(value) > max_length:
        raise ValueError(f'{error_str} is longer than maximum allowed length {max_length}')

    if regex and not re.match(regex, value):
        raise ValueError(f'{error_str} does not match regex {regex}')


class ProxyInfo(TypedDict):
    """Provides information about a proxy connection that is used for requests."""

    url: str
    """The URL of the proxy."""

    hostname: str
    """The hostname of the proxy."""

    port: int
    """The proxy port."""

    username: NotRequired[str]
    """The username for the proxy."""

    password: str
    """The password for the proxy."""

    groups: NotRequired[List[str]]
    """An array of proxy groups to be used by the [Apify Proxy](https://docs.apify.com/proxy).
    If not provided, the proxy will select the groups automatically.
    """

    country_code: NotRequired[str]
    """If set and relevant proxies are available in your Apify account, all proxied requests will
    use IP addresses that are geolocated to the specified country. For example `GB` for IPs
    from Great Britain. Note that online services often have their own rules for handling
    geolocation and thus the country selection is a best attempt at geolocation, rather than
    a guaranteed hit. This parameter is optional, by default, each proxied request is assigned
    an IP address from a random country. The country code needs to be a two letter ISO country code.
    See the [full list of available country codes](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2#Officially_assigned_code_elements).
    This parameter is optional, by default, the proxy uses all available proxy servers from all countries.
    """

    session_id: NotRequired[str]
    """The identifier of the used proxy session, if used. Using the same session ID guarantees getting the same proxy URL."""


class ProxyConfiguration:
    """Configures a connection to a proxy server with the provided options.

    Proxy servers are used to prevent target websites from blocking your crawlers based on IP address rate limits or blacklists.
    The default servers used by this class are managed by [Apify Proxy](https://docs.apify.com/proxy).
    To be able to use Apify Proxy, you need an Apify account and access to the selected proxies. If you provide no configuration option,
    the proxies will be managed automatically using a smart algorithm.

    If you want to use your own proxies, use the `proxy_urls` or `new_url_function` constructor options.
    Your list of proxy URLs will be rotated by the configuration, if this option is provided.

    **Example usage:**

    ```python

    proxy_configuration = await Actor.create_proxy_configuration(
      groups=['GROUP1', 'GROUP2']  # List of Apify Proxy groups
      country_code='US',
    )

    proxy_url = await proxy_configuration.new_url()

    with httpx.AsyncClient(proxies=proxy_url) as client:
        response = await client.get(...)
    ```
    """

    is_man_in_the_middle = False

    _next_custom_url_index = 0
    _proxy_urls: List[str]
    _used_proxy_urls: Dict[str, str]
    _new_url_function: Optional[Union[Callable[[Optional[str]], str], Callable[[Optional[str]], Awaitable[str]]]] = None
    _groups: List[str]
    _country_code: Optional[str] = None
    _password: Optional[str] = None
    _hostname: str
    _port: int
    _uses_apify_proxy: Optional[bool] = None
    _actor_config: Configuration
    _apify_client: Optional[ApifyClientAsync] = None

    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def __init__(
        self,
        *,
        password: Optional[str] = None,
        groups: Optional[List[Annotated[str, Field(regex=APIFY_PROXY_VALUE_REGEX)]]] = None,
        country_code: Optional[Annotated[str, Field(regex=COUNTRY_CODE_REGEX)]] = None,
        proxy_urls: Optional[List[Annotated[str, AnyUrl]]] = None,
        new_url_function: Optional[Union[Callable[[Optional[str]], str], Callable[[Optional[str]], Awaitable[str]]]] = None,
        actor_config: Optional[Configuration] = None,
        apify_client: Optional[ApifyClientAsync] = None,
    ):
        """Create a ProxyConfiguration instance. It is highly recommended to use `Actor.create_proxy_configuration()` instead of this.

        Args:
            password (str, optional): Password for the Apify Proxy. If not provided, will use os.environ['APIFY_PROXY_PASSWORD'], if available.
            groups (list of str, optional): Proxy groups which the Apify Proxy should use, if provided.
            country_code (str, optional): Country which the Apify Proxy should use, if provided.
            proxy_urls (list of str, optional): Custom proxy server URLs which should be rotated through.
            new_url_function (Callable, optional): Function which returns a custom proxy URL to be used.
        """  # noqa: D417 # we don't want to document the actor_config and apify_client arguments, they're internal
        if groups:
            for group in groups:
                _check(group, label='groups', regex=APIFY_PROXY_VALUE_REGEX)
        if country_code:
            _check(country_code, label='country_code', regex=COUNTRY_CODE_REGEX)

        # Validation
        if proxy_urls and new_url_function:
            raise ValueError('Cannot combine custom proxies in "proxy_urls" with custom generating function in "new_url_function".')

        if (proxy_urls or new_url_function) and (groups or country_code):
            raise ValueError('Cannot combine custom proxies with Apify Proxy!'
                             ' It is not allowed to set "proxy_urls" or "new_url_function" combined with'
                             ' "groups" or "country_code".')

        # mypy has a bug with narrowing types for filter (https://github.com/python/mypy/issues/12682)
        if proxy_urls and next(filter(lambda url: 'apify.com' in url, proxy_urls), None):  # type: ignore
            print('Some Apify proxy features may work incorrectly. Please consider setting up Apify properties instead of `proxy_urls`.\n'
                  'See https://sdk.apify.com/docs/guides/proxy-management#apify-proxy-configuration')

        self._actor_config = actor_config or Configuration._get_default_instance()
        self._apify_client = apify_client

        self._hostname = self._actor_config.proxy_hostname
        self._port = self._actor_config.proxy_port
        self._password = password or self._actor_config.proxy_password

        self._proxy_urls = list(proxy_urls) if proxy_urls else []
        self._used_proxy_urls = {}
        self._new_url_function = new_url_function
        self._groups = list(groups) if groups else []
        self._country_code = country_code
        self._uses_apify_proxy = not (proxy_urls or new_url_function)

    async def initialize(self) -> None:
        """Load the Apify Proxy password if the API token is provided and check access to Apify Proxy and provided proxy groups.

        Only called if Apify Proxy configuration is used.
        Also checks if country has access to Apify Proxy groups if the country code is provided.

        You should use the Actor.create_proxy_configuration function
        to create a pre-initialized `ProxyConfiguration` instance instead of calling this manually.
        """
        if self._uses_apify_proxy:
            await self._maybe_fetch_password()
            await self._check_access()

    @validate_arguments
    async def new_url(self, session_id: Optional[Annotated[Union[int, str], SESSION_ID_PYDANTIC_FIELD]] = None) -> str:
        """Return a new proxy URL based on provided configuration options and the `sessionId` parameter.

        Args:
            session_id (int or str, optional): Represents the identifier of a proxy session (https://docs.apify.com/proxy#sessions).
            All the HTTP requests going through the proxy with the same session identifier
            will use the same target proxy server (i.e. the same IP address).
            The identifier must not be longer than 50 characters and include only the following: `0-9`, `a-z`, `A-Z`, `"."`, `"_"` and `"~"`.

        Returns:
            str: A string with a proxy URL, including authentication credentials and port number.
                 For example, `http://bob:password123@proxy.example.com:8000`
        """
        if session_id is not None:
            session_id = f'{session_id}'
            _check(session_id, label='session_id', max_length=SESSION_ID_MAX_LENGTH, regex=APIFY_PROXY_VALUE_REGEX)

        if self._new_url_function:
            try:
                res = self._new_url_function(session_id)
                if inspect.isawaitable(res):
                    res = await res
                return str(res)
            except Exception as e:
                raise ValueError('The provided "new_url_function" did not return a valid URL') from e

        if self._proxy_urls:
            if not session_id:
                index = self._next_custom_url_index
                self._next_custom_url_index = (self._next_custom_url_index + 1) % len(self._proxy_urls)
                return self._proxy_urls[index]
            else:
                if session_id not in self._used_proxy_urls:
                    index = self._next_custom_url_index
                    self._next_custom_url_index = (self._next_custom_url_index + 1) % len(self._proxy_urls)
                    self._used_proxy_urls[session_id] = self._proxy_urls[index]

                return self._used_proxy_urls[session_id]

        username = self._get_username(session_id)

        return f'http://{username}:{self._password}@{self._hostname}:{self._port}'

    @validate_arguments
    async def new_proxy_info(self, session_id: Optional[Annotated[Union[int, str], SESSION_ID_PYDANTIC_FIELD]] = None) -> ProxyInfo:
        """Create a new ProxyInfo object.

        Use it if you want to work with a rich representation of a proxy URL.
        If you need the URL string only, use `ProxyConfiguration.new_url`.

        Args:
            session_id (int or str, optional): Represents the identifier of a proxy session (https://docs.apify.com/proxy#sessions).
            All the HTTP requests going through the proxy with the same session identifier
            will use the same target proxy server (i.e. the same IP address).
            The identifier must not be longer than 50 characters and include only the following: `0-9`, `a-z`, `A-Z`, `"."`, `"_"` and `"~"`.

        Returns:
            ProxyInfo: Dictionary that represents information about the proxy and its configuration.
        """
        if session_id is not None:
            session_id = f'{session_id}'
            _check(session_id, label='session_id', max_length=SESSION_ID_MAX_LENGTH, regex=APIFY_PROXY_VALUE_REGEX)

        url = await self.new_url(session_id)
        res: ProxyInfo
        if self._uses_apify_proxy:
            res = {
                'url': url,
                'hostname': self._hostname,
                'port': self._port,
                'username': self._get_username(session_id),
                'password': self._password or '',
                'groups': self._groups,
            }
            if self._country_code:
                res['country_code'] = self._country_code
            if session_id is not None:
                res['session_id'] = session_id
            return res
        else:
            parsed_url = urlparse(url)
            assert parsed_url.hostname is not None
            assert parsed_url.port is not None
            res = {
                'url': url,
                'hostname': parsed_url.hostname,
                'port': parsed_url.port,
                'password': parsed_url.password or '',
            }
            if parsed_url.username:
                res['username'] = parsed_url.username
        return res

    async def _maybe_fetch_password(self) -> None:
        token = self._actor_config.token

        if token and self._apify_client:
            user_info = await self._apify_client.user().get()
            if user_info:
                password = user_info['proxy']['password']

                if self._password:
                    if self._password != password:
                        print('The Apify Proxy password you provided belongs to'
                              ' a different user than the Apify token you are using. Are you sure this is correct?')
                else:
                    self._password = password

        if not self._password:
            raise ValueError('Apify Proxy password must be provided using the "password" constructor argument'
                             f' or the "{ApifyEnvVars.PROXY_PASSWORD}" environment variable.'
                             f' If you add the "{ApifyEnvVars.TOKEN}" environment variable, the password will be automatically inferred.')

    async def _check_access(self) -> None:
        proxy_status_url = f'{self._actor_config.proxy_status_url}/?format=json'

        status = None
        async with httpx.AsyncClient(proxies=await self.new_url()) as client:
            for _ in range(2):
                try:
                    response = await client.get(proxy_status_url)
                    status = response.json()
                    break
                except Exception:
                    # retry on connection errors
                    pass

        if status:
            if not status['connected']:
                raise ConnectionError(status['connectionError'])

            self.is_man_in_the_middle = status['isManInTheMiddle']
        else:
            print('Apify Proxy access check timed out. Watch out for errors with status code 407. '
                  "If you see some, it most likely means you don't have access to either all or some of the proxies you're trying to use.")

    def _get_username(self, session_id: Optional[Union[int, str]] = None) -> str:
        if session_id is not None:
            session_id = f'{session_id}'

        parts: List[str] = []

        if self._groups:
            parts.append(f'groups-{"+".join(self._groups)}')
        if session_id is not None:
            parts.append(f'session-{session_id}')
        if (self._country_code):
            parts.append(f'country-{self._country_code}')

        if not parts:
            return 'auto'

        return ','.join(parts)



import asyncio
import inspect
import json
import traceback
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Set

import websockets.client
from pyee.asyncio import AsyncIOEventEmitter

from ._utils import _maybe_extract_enum_member_value
from .config import Configuration
from .consts import ActorEventType


class EventManager:
    """A class for managing actor events.

    You shouldn't use this class directly,
    but instead use it via the `Actor.on()` and `Actor.off()` methods.
    """

    _platform_events_websocket: Optional[websockets.client.WebSocketClientProtocol] = None
    _process_platform_messages_task: Optional[asyncio.Task] = None
    _send_persist_state_interval_task: Optional[asyncio.Task] = None
    _send_system_info_interval_task: Optional[asyncio.Task] = None
    _listener_tasks: Set[asyncio.Task]
    _listeners_to_wrappers: Dict[ActorEventType, Dict[Callable, List[Callable]]]

    def __init__(self, config: Configuration) -> None:
        """Crate an instance of EventManager.

        Args:
            config (Configuration): The actor configuration to be used in this event manager.
        """
        self._config = config
        self._event_emitter = AsyncIOEventEmitter()
        self._initialized = False
        self._listener_tasks = set()
        self._listeners_to_wrappers = defaultdict(lambda: defaultdict(list))

    async def init(self) -> None:
        """Initialize the event manager.

        When running this on the Apify Platform, this will start processing events
        send by the platform to the events websocket and emitting them as events
        that can be listened to by the `Actor.on()` method.
        """
        if self._initialized:
            raise RuntimeError('EventManager was already initialized!')

        # Run tasks but don't await them
        if self._config.actor_events_ws_url:
            self._process_platform_messages_task = asyncio.create_task(self._process_platform_messages())
        else:
            print('ACTOR_EVENTS_WS_URL env var not set, no events from Apify platform will be emitted.')

        self._initialized = True

    async def close(self, event_listeners_timeout_secs: Optional[int] = None) -> None:
        """Initialize the event manager.

        This will stop listening for the platform events,
        and it will wait for all the event listeners to finish.

        Args:
            event_listeners_timeout_secs (int, optional): Optional timeout after which the pending event listeners are canceled.
        """
        if not self._initialized:
            raise RuntimeError('EventManager was not initialized!')

        if self._platform_events_websocket:
            await self._platform_events_websocket.close()

        if self._process_platform_messages_task:
            await self._process_platform_messages_task

        await self.wait_for_all_listeners_to_complete(timeout_secs=event_listeners_timeout_secs)

        self._event_emitter.remove_all_listeners()

        self._initialized = False

    def on(self, event_name: ActorEventType, listener: Callable) -> Callable:
        """Add an event listener to the event manager.

        Args:
            event_name (ActorEventType): The actor event for which to listen to.
            listener (Callable): The function which is to be called when the event is emitted (can be async).
        """
        if not self._initialized:
            raise RuntimeError('EventManager was not initialized!')

        event_name = _maybe_extract_enum_member_value(event_name)

        async def inner_wrapper(*args: Any, **kwargs: Any) -> None:
            if inspect.iscoroutinefunction(listener):
                await listener(*args, **kwargs)
            else:
                listener(*args, **kwargs)

        async def outer_wrapper(*args: Any, **kwargs: Any) -> None:
            listener_task = asyncio.create_task(inner_wrapper(*args, **kwargs))
            self._listener_tasks.add(listener_task)
            try:
                await listener_task
            finally:
                self._listener_tasks.remove(listener_task)

        self._listeners_to_wrappers[event_name][listener].append(outer_wrapper)

        return self._event_emitter.add_listener(event_name, outer_wrapper)

    def off(self, event_name: ActorEventType, listener: Optional[Callable] = None) -> None:
        """Remove a listener, or all listeners, from an actor event.

        Args:
            event_name (ActorEventType): The actor event for which to remove listeners.
            listener (Callable, optional): The listener which is supposed to be removed. If not passed, all listeners of this event are removed.
        """
        if not self._initialized:
            raise RuntimeError('EventManager was not initialized!')

        event_name = _maybe_extract_enum_member_value(event_name)

        if listener:
            for listener_wrapper in self._listeners_to_wrappers[event_name][listener]:
                self._event_emitter.remove_listener(event_name, listener_wrapper)
            self._listeners_to_wrappers[event_name][listener] = []
        else:
            self._listeners_to_wrappers[event_name] = defaultdict(list)
            self._event_emitter.remove_all_listeners(event_name)

    def emit(self, event_name: ActorEventType, data: Any) -> None:
        """Emit an actor event manually.

        Args:
            event_name (ActorEventType): The actor event which should be emitted.
            data (Any): The data that should be emitted with the event.
        """
        event_name = _maybe_extract_enum_member_value(event_name)

        self._event_emitter.emit(event_name, data)

    async def wait_for_all_listeners_to_complete(self, *, timeout_secs: Optional[int] = None) -> None:
        """Wait for all event listeners which are currently being executed to complete.

        Args:
            timeout_secs (int, optional): Timeout for the wait. If the event listeners don't finish until the timeout, they will be canceled.
        """
        async def _wait_for_listeners() -> None:
            results = await asyncio.gather(*self._listener_tasks, return_exceptions=True)
            for result in results:
                if result is Exception:
                    print('Exception in one of the event listeners:')
                    traceback.print_exception(type(result), result, result.__traceback__)

        if timeout_secs:
            _, pending = await asyncio.wait([asyncio.create_task(_wait_for_listeners())], timeout=timeout_secs)
            if pending:
                print('Timed out waiting for event listeners to complete, unfinished event listeners will be canceled')
                for pending_task in pending:
                    pending_task.cancel()
                    try:
                        await pending_task
                    except asyncio.CancelledError:
                        pass
        else:
            await _wait_for_listeners()

    async def _process_platform_messages(self) -> None:
        # This should be called only on the platform, where we have the ACTOR_EVENTS_WS_URL configured
        assert self._config.actor_events_ws_url is not None

        try:
            async with websockets.client.connect(self._config.actor_events_ws_url) as websocket:
                self._platform_events_websocket = websocket
                async for message in websocket:
                    try:
                        parsed_message = json.loads(message)
                        event_name = parsed_message['name']
                        event_data = parsed_message['data']

                        self._event_emitter.emit(event_name, event_data)

                    except Exception as e:
                        print('Cannot parse actor event', e)
        except Exception as e:
            print('Error in websocket connection', e)
